# cython: language_level=3
# no-maintain-checks
"""
DICT(IONARY) (DATA)SET

A class creating a Data Frame type construct with lists of dictionaries.

(C) 2021 Justin Joyce.

https://github.com/joocer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
# python setup.py build_ext --inplace

import os
import statistics
from enum import Enum
from functools import reduce
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Union

import orjson
from orso.cityhash import CityHash64

from mabel.data.internals.display import ascii_table
from mabel.data.internals.display import html_table
from mabel.data.internals.dnf_filters import DnfFilters
from mabel.data.internals.dumb_iterator import DumbIterator
from mabel.data.internals.expression import Expression
from mabel.data.internals.group_by import GroupBy
from mabel.data.internals.storage_classes import StorageClassCompressedMemory
from mabel.data.internals.storage_classes import StorageClassDisk
from mabel.data.internals.storage_classes import StorageClassMemory
from mabel.errors import MissingDependencyError
from mabel.utils.ipython import is_running_from_ipython


class STORAGE_CLASS(int, Enum):
    """
    How to cache the results for processing:

    - NO_PERSISTANCE = don't do anything with the records to cache them, assumes
      the records are already persisted (e.g. you've loaded a list) but most
      functionality works with generators.
    - MEMORY = load the entire dataset into a list, this is fast but if the
      dataset is too large it will kill the app.
    - DISK = load the entire dataset to a temporary file, this can deal with
      Tb of data (if you have that much disk space) but it is at least 3x slower
      than memory from basic bench marks.
    - COMPRESSED_MEMORY = a middle ground, allows you to load more data into
      memory but still needs to perform compression on the data so isn't as fast
      as the MEMORY option. Bench marks show you can fit about 2x the data in
      memory but at a cost of 2.5x - your results will vary.
    """

    NO_PERSISTANCE = 1
    MEMORY = 2
    DISK = 3
    COMPRESSED_MEMORY = 4


class DictSet(object):
    def __init__(
        self,
        iterator: Iterable[Dict[Any, Any]],
        *,
        storage_class=STORAGE_CLASS.NO_PERSISTANCE,
    ):
        """
        Create a DictSet.

        Parameters:
            iterator: Iterable
                An iterable which is our DictSet
            persistance: STORAGE_CLASS (optional)
                How to store this dataset while we're processing it. The default is
                NO_PERSISTANCE which applies no specific persistance. MEMORY loads
                into a Python `list`, DISK saves to disk - disk persistance is slower
                but can handle much larger data sets. 'COMPRESSED_MEMORY' uses
                compression to fit more in memory for a performance cost.
        """
        self.storage_class = storage_class
        self._iterator = iterator
        self._temporary_folder = None

        # if we're persisting to memory, load into a list
        if storage_class == STORAGE_CLASS.MEMORY:
            self._iterator = StorageClassMemory(iterator)  # type:ignore

        # if we're persisting to disk, save it
        if storage_class == STORAGE_CLASS.DISK:
            self._iterator = StorageClassDisk(iterator)

        # if we're persisiting to compressed memory, do it
        if storage_class == STORAGE_CLASS.COMPRESSED_MEMORY:
            self._iterator = StorageClassCompressedMemory(iterator)

        if not hasattr(self._iterator, "__iter__"):  # pragma:no cover
            self._iterator = DumbIterator(self._iterator)

    def __iter__(self):
        """
        Wrap the iterator in a Iterable object
        """
        if not hasattr(self._iterator, "__iter__"):  # pragma:no cover
            self._iterator = DumbIterator(self._iterator)
        return self

    def __next__(self):
        return next(self._iterator)

    def __enter__(self):  # pragma:no cover
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pragma:no cover
        pass  # exist needs to exist to be a context manager

    def sample(self, fraction: float = 0.5):
        """
        Select a random sample of records, fraction indicates the portion of
        records to select.

        NOTE: records are randomly selected so is unlikely to perfectly match the
        fraction.
        """

        def inner_sampler(dictset):
            selector = int(1 / fraction)
            for row in dictset:
                random_value = int.from_bytes(os.urandom(2), "big")
                if random_value % selector == 0:
                    yield row

        return DictSet(inner_sampler(iter(self._iterator)), storage_class=self.storage_class)

    def collect_list(self, key: str = None) -> Union[list, map]:
        """
        Convert a _DictSet_ to a list, optionally, but probably usually, just extract
        a specific column.

        Return None if the value in the field is None, if the field doesn't exist in
        the record, don't return anything.
        """

        def asdic(lst):
            for l in lst:
                if isinstance(l, dict):
                    yield l
                else:
                    yield l.as_dict()

        if not key:
            return list(asdic(iter(self._iterator)))
        return [record[key] for record in iter(self._iterator) if key in record]

    def keys(self, number_of_rows: int = 0):
        """
        Get all of the keys from the _DictSet_. This iterates the entire
        _DictSet_ unless told not to.
        """
        if number_of_rows > 0:
            rows = self.itake(number_of_rows)
            return reduce(lambda x, y: x + [a for a in y.keys() if a not in x], rows, [])
        return reduce(
            lambda x, y: x + [a for a in y.keys() if a not in x],
            iter(self._iterator),
            [],
        )

    def types(self, number_of_rows: int = 100):
        top = self.take(number_of_rows)
        response = {}
        for key in top.keys():
            key_type = {type(val).__name__ for val in top.collect_list(key) if val != None}
            if len(key_type) == 0:  # pragma: no cover
                response[key] = "empty"
            if len(key_type) == 1:
                response[key] = key_type.pop()
            elif sorted(key_type) == ["float", "int"]:
                response[key] = "numeric"
            else:  # pragma: no cover
                response[key] = "mixed"
        return response

    def max(self, key: str):
        """
        Find the maximum in a column of this _DictSet_.

        Parameters:
            key: string
                The column to perform the function on
        """
        return reduce(max, self.collect_list(key))

    def sum(self, key: str):
        """
        Find the sum of a column of this _DictSet_.

        Parameters:
            key: string
                The column to perform the function on
        """
        return reduce(lambda x, y: x + y, self.collect_list(key), 0)

    def min(self, key: str):
        """
        Find the minimum in a column of this _DictSet_.

        Parameters:
            key: string
                The column to perform the function on
        """
        return reduce(min, self.collect_list(key))

    def min_max(self, key: str):
        """
        Find the minimum and maximum of a column at the same time.

        Parameters:
            key: string
                The column to perform the function on

        Returns:
            tuple (minimum, maximum)
        """

        def minmax(a, b):
            return min(a[0], b[0]), max(a[1], b[1])

        return reduce(minmax, map(lambda x: (x, x), self.collect_list(key)))

    def mean(self, key: str):
        """
        Find the mean in a column of this _DictSet_.

        Parameters:
            key: string
                The column to perform the function on
        """
        return statistics.mean(self.collect_list(key))

    def variance(self, key: str):
        """
        Find the variance in a column of this _DictSet_.

        Parameters:
            key: string
                The column to perform the function on
        """
        return statistics.variance(self.collect_list(key))

    def standard_deviation(self, key: str):
        """
        Find the standard deviation in a column of this _DictSet_.

        Parameters:
            key: string
                The column to perform the function on
        """
        return statistics.stdev(self.collect_list(key))

    def count(self):
        """
        Count the number of items in the _DictSet_.
        """
        if hasattr(self._iterator, "__len__"):
            return len(self._iterator)
        else:
            # we can't count the items in an non persisted DictSet
            return -1

    def distinct(self, *columns):
        """
        Remove duplicates from a _DictSet_. This creates a list of the items
        already added to the result, so is not suitable for huge _DictSets_.

        Optionally accepts a list of columns, which we extract out and just
        'distinct' on these, ignoring differences in any of the other columns.
        """
        hash_list = {}

        def do_dedupe(data):
            for item in data:
                if columns:
                    hashed_item = hash("".join([str(item.get(c, "$$")) for c in columns]))
                else:
                    hashed_item = reduce(
                        lambda x, y: x ^ y,
                        [hash(f"{i},{v}") for i, v in enumerate(item.values())],
                        0,
                    )
                if hashed_item not in hash_list:
                    yield item
                    hash_list[hashed_item] = True

        return DictSet(do_dedupe(iter(self._iterator)), storage_class=self.storage_class)

    def group_by(self, group_by_columns):
        """
        Group a dictset by a column or group of columns. Returns a GroupBy object.
        """
        return GroupBy(iter(self._iterator), group_by_columns)

    def collect_set(self, column, dedupe: bool = False):
        from .collected_set import CollectedSet

        return CollectedSet(self, column, dedupe=dedupe)

    def get_items(self, *locations):
        """
        Get items from the DictSet at a set of indicies, try to find the fastest
        way possible to do this.
        """

        # if there's no direct access to items, cycle through them
        # yielding the items we want
        if self.storage_class in (STORAGE_CLASS.DISK, STORAGE_CLASS.COMPRESSED_MEMORY):
            for r in self._iterator._inner_reader(*locations):
                yield r
            return

        # if the iterator allows us to access items directly, use that
        if hasattr(self._iterator, "__getitem__"):  # pragma: no cover
            yield from [self._iterator[i] for i in locations]
            return

        if self.storage_class == STORAGE_CLASS.NO_PERSISTANCE:
            for i, r in iter(self._iterator):
                if i in locations:
                    yield r

    def to_ascii_table(self, limit: int = 5):
        """
        Return the top `limit` rows from a _DictSet_ as an ASCII table.

        Returns:
            Table encoded in a string
        """
        return ascii_table(iter(self._iterator), limit)

    def to_html_table(self, limit: int = 5):
        """
        Return the top `limit` rows from a _DictSet_ as a HTML table.

        Returns:
            HTML Table encoded in a string
        """
        return html_table(DumbIterator(self._iterator), limit)

    def to_pandas(self):
        """
        Load the contents of the _DictSet_ to a _Pandas_ DataFrame.

        Returns:
            Pandas DataFrame
        """
        try:
            import pandas
        except ImportError:  # pragma: no cover
            raise MissingDependencyError(
                "`pandas` is missing, please install or include in requirements.txt"
            )
        return pandas.DataFrame([r for r in iter(self._iterator)])

    def first(self) -> dict:
        """
        Retun the first item in the DictSet
        """
        return next(iter(self._iterator), None)

    def take(self, items: int):
        """
        Return the first _items_ number of items from the _DictSet_. This loads
        these items into memory. If returning a large number of items, use itake.
        """
        return DictSet(self.itake(items), storage_class=self.storage_class)

    def itake(self, items: int):
        """
        Return the first _items_ number of items from the _DictSet_.

        This returns a generator.
        """
        for count, item in enumerate(iter(self._iterator)):
            if isinstance(item, dict):
                yield item
            else:
                yield item.as_dict()
            # count is zero-based, if we've read the itemth item, quit
            if (count + 1) == items:
                return

    def filter(self, filters):
        """
        Filter a _DictSet_ returning only the items that match the predicate.

        Parameters:
            predicate: callable
                A function that takes a record as a parameter and should return
                False for items to be filtered
        """
        # Where clause filtering
        if isinstance(filters, str):

            def inner_filter_where(dictset):
                for record in dictset:
                    if q.evaluate(record):
                        yield record

            q = Expression(filters)
            return DictSet(
                inner_filter_where(iter(self._iterator)),
                storage_class=self.storage_class,
            )

        # DNF filtering
        if isinstance(filters, (tuple, list)):
            filter_set = DnfFilters(filters)
            return DictSet(
                DnfFilters.filter_dictset(filter_set, iter(self._iterator)),
                storage_class=self.storage_class,
            )

        # function filtering
        if hasattr(filters, "__call__"):

            def inner_filter_callable(func, dictset):
                for item in dictset:
                    if func(item):
                        yield item

            return DictSet(
                inner_filter_callable(filters, iter(self._iterator)),
                storage_class=self.storage_class,
            )

    @property
    def cursor(self):
        """
        If the DictSet supports cursors, return the cursor.
        """
        if hasattr(self._iterator, "cursor"):
            return self._iterator.cursor
        return None

    def project(self, columns):
        return self.select(columns)

    def select(self, columns):
        """
        Selects columns from a _DictSet_. If the column doesn't exist it is populated
        with `None`.
        """
        if columns == "*" or columns == ["*"]:
            return self

        if not isinstance(columns, (list, set, tuple)):
            columns = list([columns])

        if "*" in columns:

            def inner_select(it):
                for record in it:
                    yield {k: v for k, v in record.items() if k != "*"}

        else:

            def inner_select(it):
                for record in it:
                    yield {k: record.get(k, None) for k in columns}

        return DictSet(inner_select(iter(self._iterator)), storage_class=self.storage_class)

    def sort_and_take(self, column, take: int = 5000, descending: bool = False):
        def safety_key(column):
            # this returns a tuple where the first element is a boolean, and the
            # second item is the value
            return lambda x: (
                x.get(column) is not None,
                x.get(column),
            )

        if self.storage_class == STORAGE_CLASS.MEMORY:
            yield from sorted(self._iterator, key=safety_key(column), reverse=descending)[:take]

        else:
            # In a low-memory environment we probably can't store all of the records
            # into memory, but if we're only interested in, say the top 10, then we
            # only need to store about that many in memory at any one time. This
            # implementation stores double and one records in memory as it collects
            # and sorts them.
            double_cache = max(take * 2, 1) + 1
            cache = []
            for record in iter(self._iterator):
                cache.append(record)
                if len(cache) > double_cache:
                    cache.sort(key=safety_key(column), reverse=descending)
                    del cache[take:]
            cache.sort(key=safety_key(column), reverse=descending)
            yield from cache[:take]

    def __getitem__(self, columns):
        """
        Select the columns from the _DictSet_, alias for .select
        """
        return self.select(columns)

    def __hash__(self, seed: int = 703115) -> int:
        """
        Creates a consistent hash of the _DictSet_ regardless of the order of
        the items in the _DictSet_.
        """
        # The seed is the mission duration of the Apollo 11 mission.
        #   703115 = 8 days, 3 hours, 18 minutes, 35 seconds
        ordered = map(lambda record: dict(sorted(record.items())), iter(self._iterator))
        serialized = map(orjson.dumps, ordered)
        hashed = map(CityHash64, serialized)
        return reduce(lambda x, y: x ^ y, hashed, seed)

    def __repr__(self):  # pragma: no cover
        if is_running_from_ipython():
            from IPython.display import HTML  # type:ignore
            from IPython.display import display

            html = html_table(iter(self._iterator), 10)
            display(HTML(html))
            return ""  # __repr__ must return something
        else:
            return ascii_table(iter(self._iterator), 10)



"""
This is a filtering mechanism to be applied when reading data.
"""
import operator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from mabel.errors import InvalidSyntaxError
from mabel.utils.text import like
from mabel.utils.text import matches

# convert text representation of operators to functions
OPERATORS = {
    "=": operator.eq,
    "==": operator.eq,
    "is": operator.eq,
    "!=": operator.ne,
    "<>": operator.ne,
    "<": operator.lt,
    ">": operator.gt,
    "<=": operator.le,
    ">=": operator.ge,
    "like": like,
    "matches": matches,
    "~": matches,
    "in": lambda x, y: x in y,
    "!in": lambda x, y: x not in y,
    "not in": lambda x, y: x not in y,
    "contains": lambda x, y: y in x,
    "!contains": lambda x, y: y not in x,
}


def evaluate(predicate: Union[tuple, list], record: dict) -> bool:
    """
    This is the evaluation routine for the Filter class.

    Implements a DNF (Disjunctive Normal Form) interpretter. Predicates in the same
    list are joined with an AND (*all()*) and predicates in adjacent lists are joined
    with an OR (*any()*). This allows for non-trivial filters to be written with just
    _tuples_ and _lists_.

    The predicates are in _tuples_ in the form (`key`, `op`, `value`) where the `key`
    is the value looked up from the record, the `op` is the operator and the `value`
    is a literal.
    """
    # No filter doesn't filter
    if predicate is None:  # pragma: no cover
        return True

    # If we have a tuple extract out the key, operator and value and do the evaluation
    if isinstance(predicate, tuple):
        key, op, value = predicate
        record_value = record.get(key, None)
        return record_value is not None and OPERATORS[op.lower()](record_value, value)

    if isinstance(predicate, list):
        # Are all of the entries tuples?
        # We AND them together (_all_ are True)
        if all(isinstance(p, tuple) for p in predicate):
            return all(evaluate(p, record) for p in predicate)

        # Are all of the entries lists?
        # We OR them together (_any_ are True)
        if all(isinstance(p, list) for p in predicate):
            return any(evaluate(p, record) for p in predicate)

        # if we're here the structure of the filter is wrong
        raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover

    raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover


class DnfFilters:
    __slots__ = ("empty_filter", "predicates")

    def __init__(self, filters: Optional[List[Tuple[str, str, object]]] = None):
        """
        A class to supporting filtering data.

        Parameters:
            filters: list of tuples
                Each tuple has format: (`key`, `op`, `value`). When run the
                filter will extract the `key` field from the dictionary and
                compare to the `value` using the operator `op`. Multiple
                filters are treated as AND, lists of ANDs are treated as ORs.
                The supported `op` values are: `=` or `==`, `!=`, `<`, `>`,
                `<=`, `>=`, `in`, `!in` (not in), `contains`, `!contains`
                (does not contain) and `like`. If the `op` is `in` or `!in`,
                the `value` must be a collection such as a _list_, a _set_
                or a _tuple_.
                `like` performs similar to the SQL operator, `%` is a
                multi-character wildcard and `_` is a single character
                wildcard.

        Examples:
            filters = Filters([('name', '==', 'john')])
            filters = Filters([('name', '!=', 'john'),('name', '!=', 'tom')])

        """
        self.empty_filter = filters is None
        self.predicates = filters if filters else []

    def filter_dictset(self, dictset: Iterable[dict]) -> Iterable:
        """
        Tests each entry in a Iterable against the filters

        Parameters:
            dictset: iterable of dictionaries
                The dictset to process

        Yields:
            dictionary
        """
        if self.empty_filter:
            yield from dictset
        else:
            yield from (record for record in dictset if evaluate(self.predicates, record))

    def __call__(self, record) -> bool:
        return evaluate(self.predicates, record)



# no-maintain-checks
# cython: language_level=3
"""
This module is compiled, any changes to it need the following to be run before they
will be effective:

python setup.py build_ext --inplace


This implements a SQL-like query syntax to filter dictionaries based on combinations
of predicates.

The implementation is sort of an expression tree, it doesn't need to be a complete
expression tree as it's only doing boolean logic.

Derived from: https://gist.github.com/leehsueh/1290686
"""
from mabel.data.readers.internals.inline_evaluator import *
from mabel.utils.dates import parse_iso
from mabel.utils.token_labeler import OPERATORS
from mabel.utils.token_labeler import TOKENS
from mabel.utils.token_labeler import Tokenizer


class InvalidExpression(BaseException):
    pass


class TreeNode:
    __slots__ = ("token_type", "value", "left", "right", "parameters")

    def __init__(self, token_type, value):
        self.token_type = token_type
        self.value = value
        self.left = None
        self.right = None
        self.parameters = []


class Expression(object):
    tokenizer = None
    root = None

    def __init__(self, exp):
        self.tokenizer = Tokenizer(exp)
        self.parse()

    def parse(self):
        self.root = self.parse_expression()

    def parse_expression(self):
        andTerm1 = self.parse_and_term()
        while self.tokenizer.has_next() and self.tokenizer.next_token_type() == TOKENS.OR:
            self.tokenizer.next()
            andTermX = self.parse_and_term()
            andTerm = TreeNode(TOKENS.OR, None)
            andTerm.left = andTerm1
            andTerm.right = andTermX
            andTerm1 = andTerm
        return andTerm1

    def parse_and_term(self):
        condition1 = self.parse_condition()
        while self.tokenizer.has_next() and self.tokenizer.next_token_type() == TOKENS.AND:
            self.tokenizer.next()
            conditionX = self.parse_condition()
            condition = TreeNode(TOKENS.AND, None)
            condition.left = condition1
            condition.right = conditionX
            condition1 = condition
        return condition1

    def parse_condition(self):
        terminal1 = None

        if self.tokenizer.has_next() and self.tokenizer.next_token_type() == TOKENS.NOT:
            not_condition = TreeNode(self.tokenizer.next_token_type(), None)
            self.tokenizer.next()
            child_condition = self.parse_condition()
            not_condition.left = child_condition
            not_condition.right = None  # NOT is a unary operator
            return not_condition

        if self.tokenizer.has_next() and self.tokenizer.next_token_type() in (
            TOKENS.FUNCTION,
            TOKENS.AGGREGATOR,
        ):
            # we don't evaluate functions as part of this module, the function call
            # should be part of the record, e.g. record["YEAR(date_of_birth)"]
            # we we extract out the function and treat is as a VARIABLE

            collector = [self.tokenizer.next_token_value(), "("]

            self.tokenizer.next()
            if (
                not self.tokenizer.has_next
                or self.tokenizer.next_token_type() != TOKENS.LEFTPARENTHESES
            ):
                raise InvalidExpression("Functions should be followed by paranthesis")

            self.tokenizer.next()
            open_parentheses = 1

            while open_parentheses > 0:
                if not self.tokenizer.has_next():
                    break
                if self.tokenizer.next_token_type() == TOKENS.RIGHTPARENTHESES:
                    open_parentheses -= 1
                elif self.tokenizer.next_token_type() == TOKENS.LEFTPARENTHESES:
                    open_parentheses += 1
                collector.append(self.tokenizer.next())

            if open_parentheses != 0:
                raise InvalidExpression("Unbalanced parantheses")

            terminal1 = TreeNode(TOKENS.VARIABLE, "".join(collector))

        if self.tokenizer.has_next() and self.tokenizer.next_token_type() == TOKENS.LEFTPARENTHESES:
            # If we have a ( then go looking for the matching )

            self.tokenizer.next()
            expression = self.parse_expression()
            if (
                self.tokenizer.has_next()
                and self.tokenizer.next_token_type() == TOKENS.RIGHTPARENTHESES
            ):
                self.tokenizer.next()
                return expression
            raise InvalidExpression(f"`)` expected, but got `{self.tokenizer.next()}`")

        if not terminal1:
            terminal1 = self.parse_terminal()
        if self.tokenizer.has_next():
            if self.tokenizer.next_token_type() == TOKENS.OPERATOR:
                condition = TreeNode(
                    self.tokenizer.next_token_type(),
                    self.tokenizer.next_token_value().upper(),
                )
                self.tokenizer.next()
                terminal2 = self.parse_terminal()
                condition.left = terminal1
                condition.right = terminal2
                return condition
            raise InvalidExpression(f"Operator expected, but got `{self.tokenizer.next()}`")
        raise InvalidExpression("Operator expected, but got nothing")

    def parse_terminal(self):
        if self.tokenizer.has_next():
            token_type = self.tokenizer.next_token_type()
            if token_type == TOKENS.INTEGER:
                n = TreeNode(token_type, int(self.tokenizer.next()))
                return n
            if token_type == TOKENS.FLOAT:
                n = TreeNode(token_type, float(self.tokenizer.next()))
                return n
            if token_type in (TOKENS.VARIABLE, TOKENS.NOT):
                n = TreeNode(token_type, self.tokenizer.next())
                return n
            if token_type == TOKENS.LITERAL:
                n = TreeNode(token_type, str(self.tokenizer.next()[1:-1]))
                return n
            if token_type == TOKENS.BOOLEAN:
                n = TreeNode(token_type, self.tokenizer.next().lower() == "true")
                return n
            if token_type == TOKENS.NULL:
                n = TreeNode(token_type, None)
                return n
            if token_type == TOKENS.DATE:
                n = TreeNode(token_type, parse_iso(self.tokenizer.next()[1:-1]))
                return n

            if token_type == TOKENS.LEFTPARENTHESES:
                collector = []
                if self.tokenizer.has_next():
                    self.tokenizer.next()
                while (
                    self.tokenizer.has_next()
                    and self.tokenizer.next_token_type() != TOKENS.RIGHTPARENTHESES
                ):
                    if self.tokenizer.next_token_type() != TOKENS.COMMA:
                        collector.append(self.interpret_value(self.tokenizer.next()))
                    else:
                        self.tokenizer.next()
                if self.tokenizer.next_token_type() == TOKENS.RIGHTPARENTHESES:
                    self.tokenizer.next()
                n = TreeNode(TOKENS.LITERAL, collector)
                return n

        raise InvalidExpression(f"Unexpected token, got `{self.tokenizer.next()}`")

    def interpret_value(self, value):
        if not isinstance(value, str):
            return value
        if value.upper() in ("TRUE", "FALSE"):
            return value.upper() == "TRUE"

        try:
            num = int(value)
            return num
        except ValueError:
            pass

        try:
            num = float(value)
            return num
        except ValueError:
            pass

        return parse_iso(value) or value

    def evaluate(self, variable_dict):
        return self.evaluate_recursive(self.root, variable_dict)

    def __call__(self, variable_dict):
        return self.evaluate_recursive(self.root, variable_dict)

    def evaluate_recursive(self, treeNode, variable_dict):
        if treeNode.token_type in (
            TOKENS.INTEGER,
            TOKENS.FLOAT,
            TOKENS.LITERAL,
            TOKENS.BOOLEAN,
            TOKENS.NULL,
            TOKENS.DATE,
        ):
            return treeNode.value
        if treeNode.token_type == TOKENS.VARIABLE:
            if treeNode.value[0] == treeNode.value[-1] == "`":
                treeNode.value = treeNode.value[1:-1]
            if treeNode.value in variable_dict:
                value = variable_dict[treeNode.value]
                return self.interpret_value(value)
            return None

        left = self.evaluate_recursive(treeNode.left, variable_dict)

        if treeNode.token_type == TOKENS.NOT:
            return not left

        right = self.evaluate_recursive(treeNode.right, variable_dict)

        if treeNode.token_type == TOKENS.OPERATOR:
            try:
                return OPERATORS[treeNode.value](left, right)
            except (TypeError, ValueError):
                return None
        if treeNode.token_type == TOKENS.AND:
            return left and right
        if treeNode.token_type == TOKENS.OR:
            return left or right

        raise InvalidExpression(f"Unexpected value of type `{str(treeNode.token_type)}`")

    def to_dnf(self):
        """
        Converting to DNF as sometimes it's easier to deal with DNF than an
        expression tree.
        """
        return self._inner_to_dnf(self.root)

    def _inner_to_dnf(self, treeNode):
        if treeNode.token_type in (
            TOKENS.INTEGER,
            TOKENS.FLOAT,
            TOKENS.BOOLEAN,
            TOKENS.NULL,
            TOKENS.VARIABLE,
        ):
            return treeNode.value

        if treeNode.token_type in (TOKENS.DATE, TOKENS.LITERAL):
            return f'"{treeNode.value}"'

        left = self._inner_to_dnf(treeNode.left)
        right = None
        if treeNode.right:
            right = self._inner_to_dnf(treeNode.right)

        if treeNode.token_type == TOKENS.AND:
            return [left, right]

        if treeNode.token_type == TOKENS.OR:
            return [[left], [right]]

        if treeNode.token_type == TOKENS.OPERATOR:
            return (left, treeNode.value, right)

        if treeNode.token_type == TOKENS.NOT:
            # this isn't strict DNF
            return ("NOT", left)



import datetime
import json
import sys
import threading
from typing import Optional

import orjson
import zstandard
from orso.logging import get_logger
from orso.schema import RelationSchema

from mabel.data.internals.records import flatten
from mabel.data.validator import schema_loader
from mabel.errors import MissingDependencyError

BLOB_SIZE = 64 * 1024 * 1024  # 64Mb, 16 files per gigabyte
SUPPORTED_FORMATS_ALGORITHMS = ("jsonl", "zstd", "parquet", "text", "flat")


def get_size(obj, seen=None):
    """
    Recursively approximate the size of objects.
    We don't know the actual size until we save, so we approximate the size based
    on some rules - this will be wrong due to RLE, headers, precision and other
    factors.
    """
    size = sys.getsizeof(obj)

    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0

    if isinstance(obj, (int, float)):
        size = 6  # probably 4 bytes, could be 8
    if isinstance(obj, bool):
        size = 1
    if isinstance(obj, (str, bytes, bytearray)):
        size = len(obj) + 4
    if obj is None:
        size = 1
    if isinstance(obj, datetime.datetime):
        size = 8

    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size = sum([get_size(v, seen) for v in obj.values()]) + 8
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen) + 8
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj]) + 8
    return size


class BlobWriter(object):
    # in som failure scenarios commit is called before __init__, so we need to define
    # this variable outside the __init__.
    buffer = bytearray()
    byte_count = 0

    def __init__(
        self,
        *,  # force params to be named
        inner_writer=None,  # type:ignore
        blob_size: int = BLOB_SIZE,
        format: str = "zstd",
        schema: Optional[RelationSchema] = None,
        **kwargs,
    ):
        self.format = format
        self.maximum_blob_size = blob_size

        if format not in SUPPORTED_FORMATS_ALGORITHMS:
            raise ValueError(
                f"Invalid format `{format}`, valid options are {SUPPORTED_FORMATS_ALGORITHMS}"
            )

        kwargs["format"] = format
        self.inner_writer = inner_writer(**kwargs)  # type:ignore

        self.open_buffer()

        if self.format == "parquet":
            self.append = self.arrow_append
        else:
            self.append = self.text_append

        self.schema = schema_loader(schema)

    def arrow_append(self, record: dict = {}):
        record_length = get_size(record)
        # if this write would exceed the blob size, close it
        if (
            self.byte_count + record_length
        ) > self.maximum_blob_size and self.records_in_buffer > 0:
            self.commit()
            self.open_buffer()

        self.byte_count += record_length + 16
        self.records_in_buffer += 1
        self.buffer.append(record)  # type:ignore

    def text_append(self, record: dict = {}):
        # serialize the record
        if self.format == "text":
            if isinstance(record, bytes):
                serialized = record + b"\n"
            elif isinstance(record, str):
                serialized = record.encode() + b"\n"
            else:
                serialized = str(record).encode() + b"\n"
        elif self.format == "flat":
            serialized = orjson.dumps(flatten(record)) + b"\n"  # type:ignore
        elif hasattr(record, "mini"):
            serialized = record.mini + b"\n"  # type:ignore
        else:
            try:
                serialized = orjson.dumps(record) + b"\n"  # type:ignore
            except TypeError:
                serialized = json.dumps(record).encode() + b"\n"

        # the newline isn't counted so add 1 to get the actual length if this write
        # would exceed the blob size, close it so another blob will be created
        if len(self.buffer) > self.maximum_blob_size and self.records_in_buffer > 0:
            self.commit()
            self.open_buffer()

        # write the record to the file
        self.buffer.extend(serialized)
        self.records_in_buffer += 1

        return self.records_in_buffer

    def _normalize_arrow_schema(self, table, mabel_schema: RelationSchema):
        """
        Because we partition the data, there are instances where nulls in one of the
        columns isn't being correctly identified as the target type.

        We only handle a subset of types here, so it doesn't remove the problem.
        """
        try:
            import pyarrow
        except ImportError:
            raise MissingDependencyError(
                "`pyarrow` missing, please install or include in `requirements.txt`."
            )

        type_map = {
            "TIMESTAMP": pyarrow.timestamp("us"),
            "DATE": pyarrow.date64(),
            "VARCHAR": pyarrow.string(),
            "BOOLEAN": pyarrow.bool_(),
            "INTEGER": pyarrow.int64(),
            "DOUBLE": pyarrow.float64(),
            "ARRAY": pyarrow.list_(pyarrow.string())
            #            "STRUCT": pyarrow.map_(pyarrow.string(), pyarrow.string())
        }

        schema = table.schema

        for column in schema.names:
            # if we know about the column and it's a type we handle
            mabel_column = mabel_schema.find_column(column)
            if mabel_column and mabel_column.type in type_map:
                index = table.column_names.index(column)
                # update the schema
                schema = schema.set(index, pyarrow.field(column, type_map[mabel_column.type]))
        # apply the updated schema
        table = table.cast(target_schema=schema)
        return table

    def commit(self):
        committed_blob_name = ""

        if len(self.buffer) > 0:
            lock = threading.Lock()

            try:
                lock.acquire(blocking=True, timeout=10)

                if self.format == "parquet":
                    try:
                        import pyarrow
                        import pyarrow.parquet
                    except ImportError as err:  # pragma: no cover
                        raise MissingDependencyError(
                            "`pyarrow` is missing, please install or include in requirements.txt"
                        )

                    import io

                    tempfile = io.BytesIO()

                    # Add in any columns from the schema
                    columns = sorted(self.schema.column_names)

                    # then we make sure each row has all the columns
                    self.buffer = [
                        {column: row.get(column) for column in columns} for row in self.buffer
                    ]

                    pytable = pyarrow.Table.from_pylist(self.buffer)

                    # if we have a schema, make effort to align the parquet file to it
                    if self.schema:
                        pytable = self._normalize_arrow_schema(pytable, self.schema)

                    pyarrow.parquet.write_table(pytable, where=tempfile, compression="zstd")

                    tempfile.seek(0)
                    self.buffer = tempfile.read()

                if self.format == "zstd":
                    # zstandard is an non-optional installed dependency
                    self.buffer = zstandard.compress(self.buffer)

                committed_blob_name = self.inner_writer.commit(
                    byte_data=bytes(self.buffer), override_blob_name=None
                )

                if "BACKOUT" in committed_blob_name:
                    get_logger().warning(
                        f"{self.records_in_buffer:n} failed records written to BACKOUT partition `{committed_blob_name}`"
                    )
                get_logger().debug(
                    {
                        "format": self.format,
                        "committed_blob": committed_blob_name,
                        "records": len(self.buffer)
                        if self.format == "parquet"
                        else self.records_in_buffer,
                        "bytes": self.byte_count if self.format == "parquet" else len(self.buffer),
                    }
                )
            finally:
                lock.release()

        self.open_buffer()
        return committed_blob_name

    def open_buffer(self):
        if self.format == "parquet":
            self.buffer = []
            self.byte_count = 5120  # parquet has headers etc
        else:
            self.buffer = bytearray()
            self.byte_count = 0
        self.records_in_buffer = 0

    def __del__(self):
        # this should never be relied on to save data
        self.commit()



# cython: language_level=3
"""
This is an contained reading pipeline, intended to be run in a thread/processes.

The processing pipeline is made of a set of steps operating on a defined and finite
dataset:

┌────────────┬────────────────────────────────────────────────────────────┐
│ Function   │ Role                                                       │
├────────────┼────────────────────────────────────────────────────────────┤
│ Pre-Filter │ Pre-filter the rows based on a subset of the filters       │
│            │                                                            │
│ Read       │ Read the raw content from the file                         │
│            │                                                            │
│ Decompress │ Convert the raw content to lined data                      │
│            │                                                            │
│ Parse      │ Interpret the lined data into dictionaries                 │
│            │                                                            │
│ Filter     │ Apply full set of row filters to the read data             │
│            │                                                            │
│ Reduce     │ Aggregate                                                  │
└────────────┴────────────────────────────────────────────────────────────┘
"""
from enum import Enum

from orso import logging

from mabel.data.internals.dnf_filters import DnfFilters
from mabel.data.internals.expression import Expression
from mabel.data.internals.records import flatten
from mabel.utils import paths

from . import decompressors
from . import parsers

logger = logging.get_logger()


def empty_list(x):
    return []


class EXTENSION_TYPE(str, Enum):
    # labels for the file extentions
    DATA = "DATA"
    CONTROL = "CONTROL"
    INDEX = "INDEX"


KNOWN_EXTENSIONS = {
    ".txt": (decompressors.block, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".json": (decompressors.block, parsers.json, EXTENSION_TYPE.DATA),
    ".zstd": (decompressors.zstd, parsers.json, EXTENSION_TYPE.DATA),
    ".lzma": (decompressors.lzma, parsers.json, EXTENSION_TYPE.DATA),
    ".zip": (decompressors.unzip, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".jsonl": (decompressors.lines, parsers.json, EXTENSION_TYPE.DATA),
    ".xml": (decompressors.block, parsers.xml, EXTENSION_TYPE.DATA),
    ".lxml": (decompressors.lines, parsers.xml, EXTENSION_TYPE.DATA),
    ".parquet": (decompressors.parquet, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".csv": (decompressors.csv, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".ignore": (empty_list, empty_list, EXTENSION_TYPE.CONTROL),
    ".complete": (empty_list, empty_list, EXTENSION_TYPE.CONTROL),
    ".idx": (empty_list, empty_list, EXTENSION_TYPE.INDEX),
}


pass_thru = lambda x: x


def no_filter(x):
    return True


def expand_nested_json(row):
    # this is really slow - on a simple read it's roughly 60% of the execution
    if hasattr(row, "items"):
        for k, v in [(k, v) for k, v in row.items()]:
            if hasattr(v, "items"):
                if hasattr(row, "as_dict"):
                    row = row.as_dict()  # only convert if we really have to
                row.update(flatten(dictionary={k: v}, separator="."))
                row.pop(k)
    return row


class ParallelReader:
    NOT_INDEXED = {-1}

    def __init__(
        self,
        reader,
        filters=no_filter,
        columns="*",
        reducer=pass_thru,
        override_format=None,
        **kwargs,
    ):
        """

        Parameters:
            reader: callable
            columns: callable
            filters: callable
            reducer: callable
            **kwargs: kwargs
        """

        # DNF form is a representation of logical expressions which it is easier
        # to apply any indicies to - we can convert Expressions to DNF but we can't
        # convert functions to DNF.
        if isinstance(filters, DnfFilters):
            self.dnf_filter = filters
        elif isinstance(filters, Expression):
            self.dnf_filter = DnfFilters(filters.to_dnf())
        else:
            self.dnf_filter = None

        # the reader gets the data from the storage platform e.g. read the file from
        # disk or download the file
        self.reader = reader

        self.columns = columns

        # this is the filter of the collected data, this can be used
        # against more operators
        self.filters = filters

        # this is aggregation and reducers for the data
        self.reducer = reducer

        # sometimes the user knows better
        self.override_format = override_format

        if self.override_format:
            self.override_format = self.override_format.lower()
            if not self.override_format[0] == ".":
                self.override_format = "." + self.override_format

    def __call__(self, blob_name, index_files):
        # print(blob_name, "in")
        try:
            if self.override_format:
                ext = self.override_format
            else:
                bucket, path, stem, ext = paths.get_parts(blob_name)

            if ext not in KNOWN_EXTENSIONS:
                return []
            decompressor, parser, file_type = KNOWN_EXTENSIONS[ext]

            # Read
            record_iterator = self.reader.read_blob(blob_name)
            # Decompress
            record_iterator = decompressor(record_iterator)
            # Parse
            record_iterator = map(parser, record_iterator)
            # Expand Nested JSON
            # record_iterator = map(expand_nested_json, record_iterator)
            # Transform
            record_iterator = map(self.columns, record_iterator)
            # Filter
            record_iterator = filter(self.filters, record_iterator)
            # Reduce
            record_iterator = self.reducer(record_iterator)
            # Yield
            yield from record_iterator
            # print(blob_name, "out")
        except Exception as e:
            import traceback

            logger.error(f"{blob_name} had an error - {e}\n{traceback.format_exc()}")
            return []



"""
StorageClassCompressedMemory is a helper class for persisting DictSets locally, it is
the backend for the COMPRESSED_MEMORY variation of the STORAGE CLASSES.

This is particularly useful when local storage is slow and the data is too big for
memory, or you are doing a lot of operations on the data.

If you're doing few operations on the the data and it is easily recreated or local
storage is fast, this isn't a good option.
"""
import gc
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

import lz4.frame

from mabel.data.internals.storage_classes import BaseStorageClass

BATCH_SIZE = 250


class StorageClassCompressedMemory(BaseStorageClass):
    def __init__(self, iterable):
        self.compressor = lz4.frame
        self.batches: List[bytearray] = []
        self.length = 0
        self.iterator = None
        self.decompressor = lz4.frame

        batch = []
        for item in iterable:
            batch.append(item)
            self.length += 1
            if len(batch) >= BATCH_SIZE:
                compressed_batch = self.compressor.compress(self.dump_json(batch))
                self.batches.append(bytearray(compressed_batch))
                batch.clear()

        if batch:
            compressed_batch = self.compressor.compress(self.dump_json(batch))
            self.batches.append(bytearray(compressed_batch))

        # Force garbage collection (though it may not provide significant benefits)
        del batch
        gc.collect()

    def _inner_reader(self, *locations: Union[int, Tuple[int], List[int], Set[int]]):
        if locations:
            ordered_location = sorted(locations)
            batch_number = -1
            batch = []
            for i in ordered_location:
                requested_batch = i // BATCH_SIZE
                if requested_batch != batch_number:
                    batch = self.parse_json(
                        self.decompressor.decompress(self.batches[requested_batch])
                    )
                yield batch[i % BATCH_SIZE]

        else:
            for batch in self.batches:
                records = self.parse_json(self.decompressor.decompress(batch))
                for record in records:
                    if record:
                        yield record

    def __iter__(self):
        self.iterator = iter(self._inner_reader())
        return self.iterator

    def __next__(self):
        if not self.iterator:
            self.iterator = iter(self._inner_reader())
        return next(self.iterator)

    def __len__(self):
        return self.length



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.errors.render_error_stack import _build_error_stack, render_error_stack
from rich import traceback

traceback.install()


def i_fail():
    raise NotImplementedError("This is a forced error to test the error stack")


def create_stack():
    variable = True
    i_fail()


def test_error_stack():
    stack = ""
    try:
        create_stack()
    except:
        stack = render_error_stack()

    # don't check the entire stack, look for key words
    assert "i_fail" in stack
    assert "This is a forced error to test the error stack" in stack, stack
    assert "test_error_stack" in stack


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



# python-dotenv allows us to create an environment file to store secrets. If
# there is no .env it will fail gracefully and fall back to the actual os
# environment.
import os
from pathlib import Path

from .data.internals.dictset import DictSet
from .data.readers.internals.sql_reader import SqlReader
from .data.readers.reader import Reader
from .data.writers.writer import Writer
from .version import __version__

try:
    import dotenv  # type:ignore
except ImportError:  # pragma: no cover
    dotenv = None  # type:ignore


env_path = Path(".") / ".env"

#  deepcode ignore PythonSameEvalBinaryExpressiontrue: false +ve, values can be different
if env_path.exists() and (dotenv is None):  # pragma: no cover  # nosemgrep
    # using logger here will tie us in knots
    print("`.env` file exists but `dotEnv` not installed.")
elif dotenv is not None:  # pragma: no cover
    dotenv.load_dotenv(dotenv_path=env_path)


if os.environ.get("RESOURCE_MONITORING", False):  # pragma: no cover
    from .utils.resource_monitoring import ResourceMonitor



"""
Writers are the target specific implementations which commit a temporary file
created by the BlobWriter to different systems, such as the filesystem,
Google Cloud Storage or MinIO.

The primary activity is contained in the .commit() method.
"""
import abc
import os
import time
import uuid
from functools import lru_cache

from mabel.utils import paths

STEM = "{stem}"


class BaseInnerWriter(abc.ABC):
    def __init__(self, **kwargs):
        dataset = kwargs.get("dataset")
        self.bucket, path, _, _ = paths.get_parts(dataset)

        if self.bucket == "/":
            self.bucket = ""
        if path == "/":
            path = ""

        self.extension = kwargs.get("extension", ".jsonl")
        if kwargs.get("format", "") in ["zstd", "parquet"]:
            self.extension = self.extension + "." + kwargs["format"]
        if kwargs.get("format") == "text":
            self.extension = ".txt"

        self.filename = self.bucket + "/" + path + STEM + self.extension
        self.filename_without_bucket = path + STEM + self.extension

    def _build_path(self):
        blob_id = f"{time.time_ns():x}-{self._get_node()}"
        return self.filename.replace(STEM, f"{blob_id}")

    @lru_cache(1)
    def _get_node(self):
        return f"{uuid.getnode():x}-{os.getpid():x}"

    @abc.abstractclassmethod
    def commit(self, byte_data, override_blob_name=None):
        pass



import decimal
from abc import ABC

import orjson


class BaseStorageClass(ABC):
    iterator = None
    length = -1

    def __init__(self, iterator):
        raise NotImplementedError()

    def _inner_reader(self, *locations):
        raise NotImplementedError()

    def __iter__(self):
        self.iterator = self._inner_reader()
        return self.iterator

    def __next__(self):
        if not self.iterator:
            self.iterator = self._inner_reader()
        return next(self.iterator)

    def __len__(self):
        return self.length

    def parse_json(self, ds):
        return orjson.loads(ds)

    def dump_json(self, ds):
        def handler(obj):
            if isinstance(obj, decimal.Decimal):
                return str(obj)
            raise TypeError

        return orjson.dumps(ds, default=handler)



"""
BigQuery Writer
"""
from enum import Enum
from typing import Optional

from google.cloud.exceptions import NotFound
from orso.logging.create_logger import get_logger

from mabel.data.writers.internals.base_inner_writer import BaseInnerWriter
from mabel.errors import MissingDependencyError

try:
    from google.cloud import bigquery  # type:ignore

    google_bigquery_installed = True
except ImportError:  # pragma: no cover
    google_bigquery_installed = False

logger = get_logger()

MILLISECONDS_PER_DAY = 86400000


class FieldMode(str, Enum):
    REQUIRED = "REQUIRED"
    NULLABLE = "NULLABLE"


FieldTypeMapping = {
    "VARCHAR": "STRING",
    "NUMERIC": "NUMERIC",
    "BOOLEAN": "BOOLEAN",
    "TIMESTAMP": "TIMESTAMP",
    "STRUCT": None,  # not supported
    "LIST": None,  # not supported
}


class GoogleBigQueryWriter(BaseInnerWriter):
    def __init__(self, project: str, credentials=None, **kwargs):
        if not google_bigquery_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`google-cloud-bigquery` is missing, please install or include in requirements.txt"
            )

        super().__init__(**kwargs)

    def _get_table(self, dataset, table, schema, partition_expiration: Optional[int] = None):
        bigquery_client = bigquery.Client()
        dataset = bigquery_client.dataset(dataset)

        # Prepares a reference to the table
        table_ref = dataset.table(table)

        try:
            table = bigquery_client.get_table(table_ref)
        except NotFound:
            # we need to work this out programatically
            schema = [
                bigquery.SchemaField(
                    name="_system_date", field_type="TIMESTAMP", mode=FieldMode.REQUIRED
                ),
                bigquery.SchemaField(name="name", field_type="STRING", mode=FieldMode.REQUIRED),
                bigquery.SchemaField(name="age", field_type="INTEGER", mode=FieldMode.REQUIRED),
            ]
            table = bigquery.Table(table_ref, schema=schema)

            if partition_expiration:
                table.time_partitioning = bigquery.TimePartitioning(
                    type_=bigquery.TimePartitioningType.DAY,
                    field="_system_date",
                    expiration_ms=partition_expiration * MILLISECONDS_PER_DAY,
                )

            table = bigquery_client.create_table(table_ref)
            logger.info(
                {
                    "action": "create",
                    "name": table.table_id,
                    "partitioned": (partition_expiration is not None),
                }
            )
        return table

    def commit(self, byte_data, override_blob_name=None):
        self.filename = self.filename_without_bucket

        # if we've been given the filename, use that, otherwise get the
        # name from the path builder
        if override_blob_name:
            blob_name = override_blob_name
        else:
            blob_name = self._build_path()

        try:
            table = self._get_table("users", "justin", None, partition_expiration=90)
        except Exception as err:
            import traceback

            logger = get_logger()
            logger.error(
                f"Error Saving Blob to GCS {type(err).__name__} - {err}\n{traceback.format_exc()}"
            )
            raise err

        return ""



"""
Base Inner Reader
"""
import abc
import datetime
import io
import pathlib
from io import IOBase
from typing import Iterable

from orso.logging import get_logger

from mabel.utils import dates
from mabel.utils import paths

BUFFER_SIZE: int = 64 * 1024 * 1024  # 64Mb


class BaseInnerReader(abc.ABC):
    def _extract_as_at(self, path):
        parts = path.split("/")
        for part in parts:
            if part.startswith("as_at_"):
                return part
        return ""

    def _extract_by(self, path):
        parts = path.split("/")
        for part in parts:
            if part.startswith("by_"):
                return part

    def __init__(self, partitions=None, partition_filter=None, **kwargs):
        today = datetime.datetime.utcnow().replace(minute=0, second=0, microsecond=0)

        self.dataset = kwargs.get("dataset")
        if self.dataset is None:
            raise ValueError("Readers must have the `dataset` parameter set")
        if not self.dataset.endswith("/"):
            self.dataset += "/"
        # if "{" not in self.dataset and not kwargs.get("raw_path", False):
        #    self.dataset += "{datefolders}/"
        if partitions:
            self.dataset += "/".join(partitions) + "/"
        self.partition_filter = partition_filter

        start_date = dates.parse_iso(kwargs.get("start_date")) or today
        end_date = dates.parse_iso(kwargs.get("end_date")) or today

        self.start_date = min(start_date, end_date)
        self.end_date = max(start_date, end_date)

        self.days_stepped_back = 0

    def step_back_a_day(self):
        """
        Steps back a day so data can be read from a previous day
        """
        self.days_stepped_back += 1
        self.start_date -= datetime.timedelta(days=1)
        self.end_date -= datetime.timedelta(days=1)
        return self.days_stepped_back

    def __del__(self):
        """
        Only here in case a helpful dev expects te base class to have it
        """
        pass

    @abc.abstractmethod
    def get_blobs_at_path(self, prefix=None) -> Iterable:
        pass

    @abc.abstractmethod
    def get_blob_bytes(self, blob: str) -> bytes:
        """
        Return a filelike object
        """
        pass

    def read_blob(self, blob: str) -> IOBase:
        """
        Read-thru cache
        """
        result = self.get_blob_bytes(blob)
        return io.BytesIO(result)

    def get_list_of_blobs(self):
        visited = {}
        blobs = []
        # For each day in the range, get the blobs for us to read
        for cycle_date in dates.date_range(self.start_date, self.end_date):
            # Build the path name
            cycle_path = pathlib.Path(paths.build_path(path=self.dataset, date=cycle_date))
            if not cycle_path in visited:
                visited[cycle_path] = True

                cycle_blobs = list(self.get_blobs_at_path(path=cycle_path))

                # Remove any BACKOUT data - this is essentially a DEAD LETTER queue
                # so we don't want to include in when reading
                cycle_blobs = [blob for blob in cycle_blobs if "BACKOUT" not in blob]

                # The partitions are stored in folders with the prefix 'by_', as in,
                # partitioned **by** field name
                list_of_partitions = {
                    self._extract_by(blob) for blob in cycle_blobs if "/by_" in blob
                }

                # If we've been provided a partition_filter search hint, try to use this
                # first to prune data
                chosen_partition = ""

                if self.partition_filter:
                    from mabel.utils import text

                    # break the filter into parts, and make sure they're safe and valid
                    (
                        partition_filter_field,
                        partition_filter_op,
                        partition_filter_value,
                    ) = self.partition_filter
                    if partition_filter_op not in ("=", "=="):
                        raise NotImplementedError(
                            "`partition_filter` operation can only be equals (`=`)"
                        )
                    partition_filter_field = text.sanitize(partition_filter_field)
                    partition_filter_value = text.sanitize(partition_filter_value)
                    partition_filter = f"/by_{partition_filter_field}/{partition_filter_field}={partition_filter_value}/"

                    # If we can find the partition in the folder set, then prune to it
                    if any([f"by_{partition_filter_field}" in by for by in list_of_partitions]):
                        # Do the pruning
                        cycle_blobs = [blob for blob in cycle_blobs if partition_filter in blob]
                        #  We only have one partition now
                        list_of_partitions = [f"by_{partition_filter_field}"]
                        get_logger().debug(f"Applied partition filter by: `{partition_filter}`")
                    else:
                        get_logger().debug(
                            f"Wasn't able to find partition to filter by: `{partition_filter}`"
                        )

                # If we have multiple 'by_' partitions, pick one (pick the first one)
                if list_of_partitions:
                    list_of_partitions = sorted(list_of_partitions)
                    chosen_partition = list_of_partitions.pop()
                    if list_of_partitions:
                        get_logger().info(
                            f"Ignoring {len(list_of_partitions)} 'by' partitionings, reading from '{chosen_partition}'"
                        )
                    # Do the pruning
                    cycle_blobs = [blob for blob in cycle_blobs if f"/{chosen_partition}/" in blob]

                def safe_get_next(lst, item):
                    try:
                        index = lst.index(item)
                        return lst[index + 1]
                    except:
                        return None

                # Cycle over the list of partitions (e.g. the hour=02 bits) we can't use
                # the frame id of one on the rest
                if chosen_partition == "":
                    partitioned_folders = {""}
                else:
                    partitioned_folders = {
                        safe_get_next(blob.split("/"), chosen_partition) for blob in cycle_blobs
                    }

                for partitioned_folder in partitioned_folders:
                    partitioned_blobs = [
                        blob
                        for blob in cycle_blobs
                        if f"{chosen_partition}/{partitioned_folder}" in blob
                    ]

                    # Work out if there's an as_at part
                    as_ats = {
                        self._extract_as_at(blob) for blob in partitioned_blobs if "as_at_" in blob
                    }
                    if as_ats:
                        as_ats = sorted(as_ats)
                        as_at = as_ats.pop()

                        is_complete = lambda blobs: any(
                            [blob for blob in blobs if as_at + "/frame.complete" in blob]
                        )
                        is_invalid = lambda blobs: any(
                            [blob for blob in blobs if (as_at + "/frame.ignore" in blob)]
                        )

                        while not is_complete(partitioned_blobs) or is_invalid(partitioned_blobs):
                            if not is_complete(partitioned_blobs):
                                get_logger().debug(
                                    f"Frame `{partitioned_folder}/{as_at}` is not complete - `frame.complete` file is not present - skipping this frame."
                                )
                            if is_invalid(partitioned_blobs):
                                get_logger().debug(
                                    f"Frame `{partitioned_folder}/{as_at}` is invalid - `frame.ignore` file is present - skipping this frame."
                                )
                            if len(as_ats) > 0:
                                as_at = as_ats.pop()
                            else:
                                as_at = None
                                break
                        if as_at is None:
                            get_logger().error(
                                f"There are no valid frames at `{partitioned_folder}`"
                            )
                            partitioned_blobs = []
                        else:
                            get_logger().debug(f"Reading from DataSet frame `{as_at}`")
                            partitioned_blobs = [
                                blob
                                for blob in partitioned_blobs
                                if (as_at in blob) and ("/frame.complete" not in blob)
                            ]

                    blobs += partitioned_blobs

        return sorted(blobs)



"""
Validator Tests
"""
import datetime
import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data.validator import schema_loader
from orso.exceptions import DataValidationError
from rich import traceback
import orjson

traceback.install()


def test_validator_all_valid_values():
    TEST_DATA = {
        "string_field": "string",
        "numeric_field": 100.0,
        "boolean_field": True,
        "dtimestamp_field": datetime.datetime.utcnow(),
        "nullable_field": None,
        "list_field": ["a", "b", "c"],
        "enum_field": "RED",
        "integer_field": 10,
        "float_field": 10.00,
        "date_field": datetime.datetime.utcnow().date(),
        "time_field": datetime.time.min,
    }
    TEST_SCHEMA = {
        "fields": [
            {"name": "string_field", "type": "VARCHAR"},
            {"name": "numeric_field", "type": "NUMERIC"},
            {"name": "boolean_field", "type": "BOOLEAN"},
            {"name": "dtimestamp_field", "type": "TIMESTAMP"},
            {"name": "nullable_field", "type": "VARCHAR"},
            {"name": "list_field", "type": "LIST"},
            {"name": "enum_field", "type": "VARCHAR", "symbols": ["RED", "GREEN", "BLUE"]},
            {"name": "integer_field", "type": "INTEGER"},
            {"name": "float_field", "type": "DOUBLE"},
            {"name": "date_field", "type": "DATE"},
            {"name": "time_field", "type": "TIME"},
        ]
    }

    test = schema_loader(TEST_SCHEMA)
    assert test.validate(TEST_DATA), test.last_error


def test_validator_invalid_string():
    TEST_DATA = {"string_field": 100}
    TEST_SCHEMA = {"fields": [{"name": "string_field", "type": "VARCHAR"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(TEST_DATA)


def test_validator_invalid_number():
    TEST_DATA = {"number_field": "one hundred"}
    TEST_SCHEMA = {"fields": [{"name": "number_field", "type": "NUMERIC"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(TEST_DATA)

    TEST_DATA = {"number_field": print}
    TEST_SCHEMA = {"fields": [{"name": "number_field", "type": "NUMERIC"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(TEST_DATA)

    TEST_DATA = {"number_field": 100.00}
    TEST_SCHEMA = {"fields": [{"name": "number_field", "type": "INTEGER"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(TEST_DATA)

    TEST_DATA = {"number_field": 100}
    TEST_SCHEMA = {"fields": [{"name": "number_field", "type": "DOUBLE"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(TEST_DATA)


def test_validator_invalid_schema():
    result = True
    try:
        schema_loader({"name": "string"})
    except:  # pragma: no cover
        result = False
    assert not result


def test_validator_invalid_boolean():
    TEST_DATA = {"boolean_field": "not true"}
    TEST_SCHEMA = {"fields": [{"name": "boolean_field", "type": "BOOLEAN"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(TEST_DATA)


def test_validator_nonnative_types():
    TEST_DATA = {
        "numeric_field": "100",
        "integer_field": "100",
        "float_field": "100.00",
        "boolean_field": "True",
        "date_field": "2000-01-01 00:00:00.0000",
        "date_field2": "2022-02-16T23:27:08.892Z",
        "nullable_field": "",
    }
    TEST_SCHEMA = {
        "fields": [
            {"name": "numeric_field", "type": "NUMERIC"},
            {"name": "integer_field", "type": "INTEGER"},
            {"name": "float_field", "type": "DOUBLE"},
            {"name": "boolean_field", "type": "BOOLEAN"},
            {"name": "date_field", "type": "TIMESTAMP"},
            {"name": "date_field2", "type": "TIMESTAMP"},
            {"name": "nullable_field", "type": "VARCHAR"},
        ]
    }

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(TEST_DATA), test.last_error


def test_validator_extended_schema():
    """
    Ensure the validator will ignore additional fields in the schema
    """
    TEST_DATA = {"string_field": "the"}
    TEST_SCHEMA = {
        "table": "this is a test schema",
        "fields": [
            {
                "name": "string_field",
                "type": "VARCHAR",
                "description": "character array",
                "last_updated": datetime.datetime.today(),
            }
        ],
    }

    test = schema_loader(TEST_SCHEMA)
    assert test.validate(TEST_DATA)


def test_validator_loaders():
    """
    Ensure dictionary, json and json files load
    """

    TEST_SCHEMA_DICT = {"fields": [{"name": "string_field", "type": "VARCHAR"}]}
    TEST_SCHEMA_STRING = orjson.dumps(TEST_SCHEMA_DICT).decode()
    TEST_SCHEMA_FILE = "temp"

    with open(TEST_SCHEMA_FILE, "w") as file:
        file.write(TEST_SCHEMA_STRING)

    failed = False
    try:
        test = schema_loader(TEST_SCHEMA_DICT)
        test.validate({"string_field": "pass"})
    except DataValidationError:  # pragma: no cover
        failed = True
    assert not failed, "load schema from dictionary"

    failed = False
    try:
        test = schema_loader(TEST_SCHEMA_STRING)
        test.validate({"string_field": "pass"})
    except DataValidationError:  # pragma: no cover
        failed = True
    assert not failed, "load schema from string"

    failed = False
    try:
        test = schema_loader(TEST_SCHEMA_FILE)
        test.validate({"string_field": "pass"})
    except DataValidationError:  # pragma: no cover
        failed = True
    assert not failed, "load schema from file"


def test_validator_list():
    INVALID_TEST_DATA = {"key": "not a list"}
    VALID_TEST_DATA = {"key": ["is", "a", "list"]}
    TEST_SCHEMA = {"fields": [{"name": "key", "type": "LIST"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(INVALID_TEST_DATA)
    assert test.validate(VALID_TEST_DATA)


def test_validator_datetime():
    INVALID_TEST_DATA_1 = {"key": "tomorrow"}
    INVALID_TEST_DATA_2 = {"key": "2020001001"}
    INVALID_TEST_DATA_3 = {"key": "2020-00-01"}
    VALID_TEST_DATA = {"key": datetime.datetime.utcnow()}
    TEST_SCHEMA = {"fields": [{"name": "key", "type": "TIMESTAMP"}]}

    test = schema_loader(TEST_SCHEMA)
    with pytest.raises(DataValidationError):
        test.validate(INVALID_TEST_DATA_1)
    with pytest.raises(DataValidationError):
        test.validate(INVALID_TEST_DATA_2)
    with pytest.raises(DataValidationError):
        test.validate(INVALID_TEST_DATA_3)
    assert test.validate(VALID_TEST_DATA)


def test_unknown_type():
    TEST_SCHEMA = {"fields": [{"name": "key", "type": "not_a_known_type"}]}

    failed = False
    try:
        test = schema_loader(TEST_SCHEMA)
    except ValueError:  # pragma: no cover
        failed = True

    assert failed


def test_validator_other():
    TEST_DATA = {"list_of_structs": [{"a": "b"}]}
    SCHEMA_LISTS = {"fields": [{"name": "list_of_structs", "type": "LIST"}]}
    #    SCHEMA_OTHER = {"fields": [{"name": "list_of_structs", "type": "OTHER"}]}

    list_test = schema_loader(SCHEMA_LISTS)
    assert list_test.validate(TEST_DATA)


#    other_test = Schema(SCHEMA_OTHER)
#    assert other_test(TEST_DATA)


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    test_validator_invalid_number()
    run_tests()



from .render_error_stack import render_error_stack


class BaseException(Exception):
    def __call__(self, *args):  # pragma: no cover
        return self.__class__(*(self.args + args))


class DataNotFoundError(BaseException):
    pass


class IntegrityError(BaseException):
    pass


class InvalidArgument(BaseException):
    pass


class InvalidCombinationError(BaseException):
    pass


class InvalidDataSetError(BaseException):
    pass


class InvalidReaderConfigError(BaseException):
    pass


class InvalidSyntaxError(BaseException):
    pass


class MissingDependencyError(BaseException):
    pass


class ValidationError(BaseException):
    pass



"""
Test the mabel logger, this extends the Python logging logger.
We test that the trace method and decorators raise no errors.
"""
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from orso.logging import LEVELS, get_logger, set_log_name
from rich import traceback

traceback.install()

os.environ["PROJECT_NAME"] = ""

LOG_NAME = "TEST_SCRIPTS"
set_log_name(LOG_NAME)


def test_new_log_levels(caplog=None):
    """
    caplog is a feature of pytest that allows logs to be captured and
    inspected.

    It is passed to a test function and has an attribute .record_tuples which
    is a stack of logging messages. To read the last item, pop it off the
    stack. It is in the form of:

    log name, log level, log message
    """
    if caplog is None:  # pragma: no cover
        print("unable to test logging interactively - use pytest")
        return

    logger = get_logger()

    logger.audit("this is a sample audit")
    res = caplog.record_tuples.pop()
    assert res == (LOG_NAME, LEVELS.AUDIT, "this is a sample audit"), res

    logger.alert("this is a sample alert")
    res = caplog.record_tuples.pop()
    assert res == (LOG_NAME, LEVELS.ALERT, "this is a sample alert"), res


def test_smoke_test():
    """
    This is just a smoke test, it exercises most of the logging functionality
    it should just work.
    """
    logger = get_logger()
    logger.setLevel(LEVELS.DEBUG)
    logger.debug("debug")
    logger.warning("warn")
    logger.warning("warn")
    logger.warning("warn")
    logger.warning("warn")
    logger.warning("warned")
    logger.warning("warn")
    logger.warning("warn")
    logger.warning("warned")
    logger.warning("warn")
    logger.warning("warn")
    logger.warning("warn")
    logger.error("error")
    logger.alert("alert")
    logger.audit("audit")


def test_log_sanitizer():
    """
    caplog records the message before the formatter so can't be used to
    test sanitization of logs automatically.

    These are another set of smoke tests, these should run without
    error.
    """
    logger = get_logger()
    logger.audit({"password": "top secret 1"})
    logger.audit('{"password": "top secret 2"}')
    logger.audit("password:topsecret3")
    logger.audit(["password", "top secret 4"])
    logger.debug({"alpha": "`1`", "beta": "`2`", "gamma": "'3'", "delta": "'4'"})
    logger.debug("{'action':'list','limit':'1000'}")


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



from typing import Any
from typing import Callable
from typing import List
from typing import MutableMapping


def select_record_fields(record: dict, fields: List[str]) -> dict:
    """
    Selects a subset of fields from a dictionary. If the field is not present
    in the dictionary it defaults to None.

    Parameters:
        record: dictionary
            The dictionary to select from
        fields: list of strings
            The list of the field names to select

    Returns:
        dictionary
    """
    return {k: record.get(k, None) for k in fields}


def order(record: dict) -> dict:
    """
    Sort a dictionary by its keys.

    Parameters:
        record: dictionary
            The dictionary to sort

    Returns:
        dictionary
    """
    return dict(sorted(record.items()))


def set_value(record: dict, field_name: str, setter: Callable) -> dict:
    """
    Sets the value of a column to either a fixed value or as the result of a
    function which recieves the row as a parameter.

    Parameters:
        record: dictionary
            The dictionary to update
        field_name: string
            The field to create or update
        setter: callable or any
            A function or constant to update the field with

    Returns:
        dictionary
    """
    copy = record.copy()
    if callable(setter):
        copy[field_name] = setter(copy)
    else:
        copy[field_name] = setter
    return copy


def flatten(dictionary: MutableMapping[Any, Any], separator: str = ".", parent_key=False):
    """
    Turn a nested dictionary into a flattened dictionary

    Parameters:
        dictionary: dictionary
            The dictionary to flatten
        parent_key: boolean
            The string to prepend to dictionary's keys
        separator: string
            The string used to separate flattened keys

    Returns:
        A flattened dictionary
    """
    items = []
    for key, value in dictionary.items():
        new_key = str(parent_key) + separator + key if parent_key else key
        if hasattr(value, "items"):
            items.extend(flatten(dictionary=value, separator=separator, parent_key=new_key).items())
        elif isinstance(value, list):
            for k, v in enumerate(value):
                items.extend(flatten({str(k): v}, new_key).items())
        else:
            items.append((new_key, value))
    return dict(items)



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel import Reader, DictSet
from mabel.data import STORAGE_CLASS
from mabel.adapters.disk import DiskReader
from orso.logging import get_logger

get_logger().setLevel(5)


STORAGE_CLASSES = [
    STORAGE_CLASS.NO_PERSISTANCE,
    STORAGE_CLASS.COMPRESSED_MEMORY,
    STORAGE_CLASS.MEMORY,
    STORAGE_CLASS.DISK,
]


def get_ds(**kwargs):
    ds = Reader(inner_reader=DiskReader, dataset="tests/data/tweets", raw_path=True, **kwargs)
    return ds


def test_count():
    for storage_class in STORAGE_CLASSES:
        ds = get_ds(persistence=storage_class)
        if storage_class == STORAGE_CLASS.NO_PERSISTANCE:
            assert ds.count() == -1, f"{storage_class} {ds.count()}"
        else:
            assert ds.count() == 50, f"{storage_class} {ds.count()}"


def test_enumeration():
    for storage_class in STORAGE_CLASSES:
        ds = get_ds(persistence=storage_class)
        i = -1
        for i, r in enumerate(ds):
            pass
        assert i + 1 == 50, f"{storage_class} {i+1}"


def test_sample():
    for storage_class in STORAGE_CLASSES:
        ds = get_ds(persistence=storage_class)
        sample = ds.sample(0.02)
        assert isinstance(sample, DictSet)
        assert sample.count() < 5, sample.count()
        assert sample.storage_class == storage_class


def test_repr():
    for storage_class in STORAGE_CLASSES:
        ds = get_ds(persistence=storage_class)
        rep = repr(ds)
        assert "├" in rep, storage_class


def test_collect():
    for storage_class in STORAGE_CLASSES:
        ds = get_ds(persistence=storage_class)
        collection = ds.collect_list("username")
        assert collection.count("NBCNews") == 44, storage_class


def test_keys():
    for storage_class in STORAGE_CLASSES:
        ds = get_ds(persistence=storage_class)
        keys = ds.keys()
        assert keys == [
            "userid",
            "username",
            "user_verified",
            "followers",
            "tweet",
            "location",
            "sentiment",
            "timestamp",
        ], storage_class


def test_types():
    for storage_class in STORAGE_CLASSES:
        if storage_class != STORAGE_CLASS.NO_PERSISTANCE:
            ds = get_ds(persistence=storage_class)
            types = ds.types()
            assert types["userid"] == "int", types
            assert types["username"] == "str"
            assert types["user_verified"] == "bool"
            assert types["followers"] == "int"
            assert types["tweet"] == "str"
            assert types["location"] == "str"
            assert types["sentiment"] == "numeric"
            assert types["timestamp"] == "str"


def test_distinct():
    for storage_class in STORAGE_CLASSES:
        if storage_class != STORAGE_CLASS.NO_PERSISTANCE:
            ds = get_ds(persistence=storage_class)
            assert ds.distinct().count() == 50, storage_class
            assert ds.distinct("username").count() == 2, storage_class
            assert ds.distinct("username", "location").count() == 2, storage_class
            assert ds.distinct("sentiment").count() == 36, storage_class


def test_summary():
    data = [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 2, "value": "two", "plus1": 3},
        {"key": 3, "value": "three", "plus1": 4},
        {"key": 4, "value": "four", "plus1": 5},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)

    assert ds.min("key") == 1
    assert ds.max("key") == 4
    assert ds.min_max("key") == (
        1,
        4,
    )
    assert ds.sum("key") == 10
    assert ds.mean("key") == 2.5
    assert ds.standard_deviation("key") == 1.2909944487358056
    assert ds.variance("key") == 1.6666666666666667


def test_take():
    data = [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 2, "value": "two", "plus1": 3},
        {"key": 3, "value": "three", "plus1": 4},
        {"key": 4, "value": "four", "plus1": 5},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    assert ds.take(1).collect_list().pop() == {
        "key": 1,
        "value": "one",
        "plus1": 2,
    }, ds.take(1).collect_list()
    assert ds.take(2).count() == 2, ds.take(2).count()

    # we had a problem where spendable dictsets were taking n+1 records, but only returning n
    burnable = DictSet((a for a in ds.collect_list()), storage_class=STORAGE_CLASS.NO_PERSISTANCE)
    print(burnable.take(3))
    l = len(list(burnable.take(10)))
    assert l == 1, l


def test_items():
    data = [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 2, "value": "two", "plus1": 3},
        {"key": 3, "value": "three", "plus1": 4},
        {"key": 4, "value": "four", "plus1": 5},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.DISK)
    items = list(ds.get_items(0, 2))
    assert items == [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 3, "value": "three", "plus1": 4},
    ], items


def test_filters():
    data = [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 2, "value": "two", "plus1": 3},
        {"key": 3, "value": "three", "plus1": 4},
        {"key": 4, "value": "four", "plus1": 5},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    dnf = ds.filter(("key", "=", 1))
    assert dnf.count() == 1
    assert dnf.first() == {"key": 1, "value": "one", "plus1": 2}

    call = ds.filter(lambda x: x["key"] == 1)
    assert call.count() == 1
    assert call.first() == {"key": 1, "value": "one", "plus1": 2}


def test_hash():
    data = [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 2, "value": "two", "plus1": 3},
        {"key": 3, "value": "three", "plus1": 4},
        {"key": 4, "value": "four", "plus1": 5},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    hashval = hash(ds)
    assert hashval == 8826660322425604498, hashval


def test_sort():
    data = [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": None, "value": "two", "plus1": 3},
        {"key": 3, "value": "three", "plus1": 4},
        {"key": 4, "value": "four", "plus1": 5},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    st = list(ds.sort_and_take("key", 100))
    assert st == [
        {"key": None, "value": "two", "plus1": 3},
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 3, "value": "three", "plus1": 4},
        {"key": 4, "value": "four", "plus1": 5},
    ], st


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""

"""
import logging
import threading
import time
from queue import Empty
from queue import SimpleQueue
from typing import Iterator

import simdjson


def json(ds):
    """parse each line in the file to a dictionary"""
    json_parser = simdjson.Parser()
    return json_parser.parse(ds)


TERMINATE_SIGNAL = -1
MAXIMUM_SECONDS_PROCESSES_CAN_RUN = 600


def page_dictset(dictset: Iterator[dict], page_size: int) -> Iterator:
    """
    Enables paging through a dictset by returning a page of records at a time.
    Parameters:
        dictset: iterable of dictionaries:
            The dictset to process
        page_size: integer:
            The number of records per page
    Yields:
        dictionary
    """
    import orjson

    chunk: list = [""] * page_size
    for i, record in enumerate(dictset):
        if i > 0 and i % page_size == 0:
            yield chunk
            chunk = [""] * page_size
            if hasattr(record, "mini"):
                chunk[0] = record.mini  # type:ignore
            else:
                chunk[0] = orjson.dumps(record)
        elif hasattr(record, "mini"):
            chunk[i % page_size] = record.mini  # type:ignore
        else:
            chunk[i % page_size] = orjson.dumps(record)
    if chunk:
        yield chunk[: i % page_size]


def _inner_process(func, source_queue, reply_queue):  # pragma: no cover
    try:
        source = source_queue.get(timeout=1)
    except Empty:  # pragma: no cover
        source = TERMINATE_SIGNAL

    while source != TERMINATE_SIGNAL:
        for chunk in page_dictset(func(source, []), 5):
            reply_queue.put(chunk, timeout=30)
        reply_queue.put(b"END OF RECORDS")
        source = None
        while source is None:
            try:
                source = source_queue.get(timeout=1)
            except Empty:  # pragma: no cover
                source = None


def processed_reader(func, items_to_read, support_files):  # pragma: no cover
    process_pool = []

    slots = 8
    reply_queue = SimpleQueue()

    send_queue = SimpleQueue()
    for item_index in range(slots):
        if item_index < len(items_to_read):
            send_queue.put(items_to_read[item_index])

    for i in range(slots):
        process = threading.Thread(
            target=_inner_process,
            args=(func, send_queue, reply_queue),
        )
        process.daemon = True
        process.start()
        process_pool.append(process)

    process_start_time = time.time()
    item_index = slots

    while any({p.is_alive() for p in process_pool}):
        try:
            records = b""
            while 1:
                records = reply_queue.get_nowait()
                if records == b"END OF RECORDS":
                    break
                yield from map(json, records)
            if item_index < len(items_to_read):
                # we use this mechanism to throttle reading blobs so we
                # don't exhaust memory
                send_queue.put_nowait(items_to_read[item_index])
                item_index += 1
            else:
                send_queue.put_nowait(TERMINATE_SIGNAL)

        except Empty:  # nosec
            if time.time() - process_start_time > MAXIMUM_SECONDS_PROCESSES_CAN_RUN:
                logging.error(
                    f"Sending TERMINATE to long running multi-processed processes after {MAXIMUM_SECONDS_PROCESSES_CAN_RUN} seconds total run time"
                )
                break
        except GeneratorExit:
            logging.error("GENERATOR EXIT DETECTED")
            break

    for process in process_pool:
        process.join()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel import Reader, DictSet
from mabel.data import STORAGE_CLASS
from mabel.data.internals.display import html_table
from mabel.adapters.disk import DiskReader
from orso.logging import get_logger

get_logger().setLevel(5)


def get_ds(**kwargs):
    ds = Reader(inner_reader=DiskReader, dataset="tests/data/tweets", raw_path=True, **kwargs)
    return ds


def test_html_table():
    ds = get_ds(persistence=STORAGE_CLASS.MEMORY)
    html = html_table(ds)

    assert "<table" in html
    assert "<th" in html
    assert "<tr" in html
    assert "</table>" in html


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import shutil
import datetime
import os
import sys
import glob
import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.disk import DiskReader, DiskWriter
from mabel.data import BatchWriter
from mabel.data import Reader


def do_writer():
    shutil.rmtree("_temp", ignore_errors=True)
    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        date=datetime.datetime.utcnow().date(),
        schema=[{"name": "character", "type": "VARCHAR"}, {"name": "persona", "type": "VARCHAR"}],
    )
    for i in range(int(1e5)):
        w.append({"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn"})
        w.append({"character": "Laszlo Cravensworth", "persona": "Jackie Daytona"})
    w.finalize()


def test_writer_without_schema_jsonl():
    shutil.rmtree("_temp", ignore_errors=True)
    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        format="jsonl",
        date=datetime.datetime.utcnow().date(),
        schema=False,
    )
    for i in range(int(1e5)):
        w.append({"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn"})
        w.append({"name": "Laszlo Cravensworth", "persona": "Jackie Daytona"})
    w.finalize()


def test_writer_without_schema_parquet():
    with pytest.raises(ValueError):
        w = BatchWriter(
            inner_writer=DiskWriter,
            dataset="_temp",
            format="parquet",
            date=datetime.datetime.utcnow().date(),
            schema=False,
        )


def test_writer_without_schema_zstd():
    shutil.rmtree("_temp", ignore_errors=True)
    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        format="zstd",
        date=datetime.datetime.utcnow().date(),
        schema=False,
    )
    for i in range(int(1e5)):
        w.append({"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn"})
        w.append({"character": "Laszlo Cravensworth", "persona": "Jackie Daytona"})
    w.finalize()


def do_writer_compressed(algo):
    shutil.rmtree("_temp", ignore_errors=True)
    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        format=algo,
        schema=[{"name": "test", "type": "BOOLEAN"}],
    )
    for i in range(int(1e5)):
        w.append({"test": True})
        w.append({"test": False})
    w.finalize()
    del w


def do_writer_default():
    shutil.rmtree("_temp", ignore_errors=True)
    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        schema=[{"name": "character", "type": "VARCHAR"}, {"name": "persona", "type": "VARCHAR"}],
    )
    for i in range(int(1e5)):
        w.append({"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn"})
        w.append({"character": "Laszlo Cravensworth", "persona": "Jackie Daytona"})
    w.finalize()
    del w


def test_reader_writer():
    do_writer()

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_format_zstd():
    do_writer_compressed("zstd")

    g = glob.glob("_temp/**/*.zstd", recursive=True)
    assert len(g) > 0, g

    c = glob.glob("_temp/**/*.complete", recursive=True)
    assert len(c) == 1, c

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_format_jsonl():
    do_writer_compressed("jsonl")

    g = glob.glob("_temp/**/*.jsonl", recursive=True)
    assert len(g) > 0, g

    c = glob.glob("_temp/**/*.complete", recursive=True)
    assert len(c) == 1, c

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_format_parquet():
    do_writer_compressed("parquet")

    g = glob.glob("_temp/**/*.parquet", recursive=True)
    assert len(g) > 0, g

    c = glob.glob("_temp/**/*.complete", recursive=True)
    assert len(c) == 1, c

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_format_default():
    do_writer_default()

    g = glob.glob("_temp/**/*.zstd", recursive=True)
    assert len(g) > 0, g

    c = glob.glob("_temp/**/*.complete", recursive=True)
    len(c) == 0, c

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_complete_anyway_no_records():
    shutil.rmtree("_temp", ignore_errors=True)

    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        always_complete=True,
        schema=[{"name": "character", "type": "VARCHAR"}, {"name": "persona", "type": "VARCHAR"}],
    )
    w.finalize()

    g = glob.glob("_temp/**/*.zstd", recursive=True)
    assert len(g) == 0, g

    c = glob.glob("_temp/**/*.complete", recursive=True)
    assert len(c) == 1, c
    shutil.rmtree("_temp", ignore_errors=True)


def test_reader_writer_complete_anyway_with_records():
    shutil.rmtree("_temp", ignore_errors=True)

    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        always_complete=True,
        schema=[{"name": "character", "type": "VARCHAR"}, {"name": "persona", "type": "VARCHAR"}],
    )
    for i in range(int(1e2)):
        w.append({"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn"})
        w.append({"character": "Laszlo Cravensworth", "persona": "Jackie Daytona"})
    w.finalize()

    g = glob.glob("_temp/**/*.zstd", recursive=True)
    assert len(g) == 1, g

    c = glob.glob("_temp/**/*.complete", recursive=True)
    assert len(c) == 1, c

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200, l


def get_data():
    r = Reader(inner_reader=DiskReader, dataset="tests/data/tweets", raw_path=True)
    return r


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    test_reader_writer()
    run_tests()



# file deepcode ignore unguarded~next~call/test: test scripts only
import datetime
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data import Reader
from mabel.adapters.disk import DiskReader
from rich import traceback

traceback.install()


def test_ignore_non_complete_frames():
    """
    test we ignore frames without the complete flag
    """
    DATA_DATE = datetime.date(2021, 3, 28)
    records = Reader(
        dataset="tests/data/framed",
        inner_reader=DiskReader,
        start_date=DATA_DATE,
        end_date=DATA_DATE,
    )
    print(next(records))
    assert next(records).get("test") == 1


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
from typing import Any
from typing import Dict
from typing import List
from typing import Union

import orjson
from orso.schema import RelationSchema

__all__ = ["schema_loader"]


def schema_loader(
    definition: Union[str, List[Dict[str, Any]], dict, RelationSchema, bool]
) -> Union[RelationSchema, bool]:
    if definition is None:
        raise ValueError(
            "Writer is missing a schema, minimum schema is a list of the columns or explicitly set to 'False' is data is unschamable."
        )

    if definition is False:
        return False

    if isinstance(definition, RelationSchema):
        return definition

    if isinstance(definition, str):
        if os.path.exists(definition):  # type:ignore
            definition = orjson.loads(open(definition, mode="r").read())  # type:ignore
        else:
            definition = orjson.loads(definition)  # type:ignore

    if isinstance(definition, dict):
        if definition.get("fields"):  # type:ignore
            definition["columns"] = definition.pop("fields")  # type:ignore

    if isinstance(definition, list):
        definition = {"columns": definition}
    definition["name"] = definition.get("name", "wal")  # type:ignore

    return RelationSchema.from_dict(definition)



# no-maintain-checks
"""
This is a basic SQL parser and interpreter - we process the parts of the SQL statement
in the following order:

FROM clause
    - includes scalar functions
WHERE clause
GROUP BY clause
    - includes aggregate functions
HAVING clause
SELECT clause
    - includes AS renames
ORDER BY clause
LIMIT clause

"""

import re
from typing import Optional

from orso.logging import get_logger

from mabel.data.readers.internals.inline_evaluator import Evaluator
from mabel.data.readers.internals.inline_evaluator import get_function_name
from mabel.utils.token_labeler import TOKENS
from mabel.utils.token_labeler import Tokenizer

SQL_PARTS = [
    r"SELECT",
    r"DISTINCT",
    r"FROM",
    r"WHERE",
    r"GROUP BY",
    r"HAVING",
    r"ORDER BY",
    r"ASC",
    r"DESC",
    r"LIMIT",
]


class InvalidSqlError(Exception):
    pass


class SqlParser:
    def __init__(self, statement):
        self.select_expression: Optional[str] = None
        self.distinct: bool = False
        self.dataset: Optional[str] = None
        self.where_expression: Optional[str] = None
        self.group_by: Optional[str] = None
        self.having: Optional[str] = None
        self.order_by: Optional[str] = None
        self.order_descending: bool = False
        self.limit: Optional[int] = None

        self.parse(statement=statement)
        self.select_evaluator = Evaluator(self.select_expression)

    def __repr__(self):
        return str(
            {
                "select": self.select_expression,
                "distinct": self.distinct,
                "from": self.dataset,
                "where": self.where_expression,
                "group by": self.group_by,
                "having": self.having,
                "order by": self.order_by,
                "descending": self.order_descending,
                "limit": self.limit,
            }
        )

    def sql_parts(self, string):
        reg = re.compile(
            r"(\(|\)|,|"
            + r"|".join([r"\b" + i.replace(r" ", r"\s") + r"\b" for i in SQL_PARTS])
            + r"|\s)",
            re.IGNORECASE,
        )
        parts = reg.split(string)
        return [part for part in parts if part != ""]

    def clean_statement(self, string):
        """
        Remove carriage returns and all whitespace to single spaces
        """
        _RE_COMBINE_WHITESPACE = re.compile(r"\s+")
        return _RE_COMBINE_WHITESPACE.sub(" ", string).strip()

    def remove_comments(self, string):
        pattern = r"(\".*?\"|\'.*?\')|(/\*.*?\*/|--[^\r\n]*$)"
        # first group captures quoted strings (double or single)
        # second group captures comments (//single-line or /* multi-line */)
        regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

        def _replacer(match):
            # if the 2nd group (capturing comments) is not None,
            # it means we have captured a non-quoted (real) comment string.
            if match.group(2) is not None:
                return ""  # so we will return empty to remove the comment
            else:  # otherwise, we will return the 1st group
                return match.group(1)  # captured quoted-string

        return regex.sub(_replacer, string)

    def validate_dataset(self, dataset):
        """
        The from clause is used to address resources in data storage; to prevent
        abusing this, we whitelist valid characters.

        Note that when this information is used, it is used in a glob query and
        only resources that exist are referenced with a reader, this should prevent
        exec flaws - but may still result in IDOR. To reduce this, use the
        `valid_dataset_prefixes` parameter on the Reader.
        """
        # start with a letter
        if not dataset[0].isalpha():
            raise InvalidSqlError("Malformed FROM clause - must start with a letter.")
        # can't be attempting path traversal
        if ".." in dataset or "//" in dataset or "--" in dataset:
            raise InvalidSqlError("Malformed FROM clause - invalid repeated characters.")
        # can only contain limited character set (alpha num . / - _)
        if (
            not dataset.replace(".", "")
            .replace("/", "")
            .replace("-", "")
            .replace("_", "")
            .isalnum()
        ):
            raise InvalidSqlError("Malformed FROM clause - invalid characters.")
        return True

    def collector(self, labeler):
        collection = []
        while labeler.has_next():
            if labeler.next_token_value().upper() not in SQL_PARTS:
                collection.append(labeler.next_token_value())
                labeler.next()
            else:
                break
        return "".join(collection).strip()

    def parse(self, statement):
        # clean the string
        clean = self.remove_comments(statement)
        clean = self.clean_statement(clean)
        # split into bits by SQL keywords
        parts = self.sql_parts(clean)
        # put into a token labeller
        labeler = Tokenizer(parts)

        while labeler.has_next():
            if labeler.peek().strip() == "":
                labeler.next()
            elif labeler.peek().upper() == "SELECT":
                labeler.next()
                while labeler.has_next() and labeler.next_token_type() == TOKENS.EMPTY:
                    labeler.next()
                if labeler.next_token_value().upper() == "DISTINCT":
                    labeler.next()
                    self.distinct = True
                    while labeler.has_next() and labeler.next_token_type() == TOKENS.EMPTY:
                        labeler.next()
                self.select_expression = self.collector(labeler)
            elif labeler.peek().upper() == "FROM":
                labeler.next()
                while labeler.has_next() and labeler.next_token_type() == TOKENS.EMPTY:
                    labeler.next()

                if labeler.next_token_type() == TOKENS.LEFTPARENTHESES:
                    # we have a subquery - we're going to collect it based on matching
                    # parentheses
                    open_parentheses = 1
                    collector = []
                    labeler.next()
                    while open_parentheses > 0:
                        if not labeler.has_next():
                            break

                        if labeler.next_token_type() == TOKENS.RIGHTPARENTHESES:
                            open_parentheses -= 1
                            if open_parentheses != 0:
                                collector.append(labeler.peek())
                        elif labeler.next_token_type() == TOKENS.LEFTPARENTHESES:
                            open_parentheses += 1
                            collector.append(labeler.peek())
                        else:
                            collector.append(labeler.peek())
                        labeler.next()

                    if open_parentheses != 0:
                        raise InvalidSqlError("Malformed FROM clause - mismatched parenthesis.")

                    self.dataset = collector
                else:
                    self.dataset = self.collector(labeler).replace(".", "/")
            elif labeler.peek().upper() == "WHERE":
                labeler.next()
                self.where_expression = self.collector(labeler)
            elif labeler.peek().upper() == "GROUP BY":
                labeler.next()
                self.group_by = self.collector(labeler)
            elif labeler.peek().upper() == "HAVING":
                labeler.next()
                self.having = self.collector(labeler)
            elif labeler.peek().upper() == "ORDER BY":
                labeler.next()
                self.order_by = self.collector(labeler)
                if labeler.has_next() and labeler.next_token_value().upper() in (
                    "ASC",
                    "DESC",
                ):
                    self.order_descending = labeler.next_token_value().upper() == "DESC"
                if labeler.has_next():
                    labeler.next()
            elif labeler.peek().upper() == "LIMIT":
                labeler.next()
                self.limit = int(self.collector(labeler))

        # validate inputs (currently only the FROM clause)
        if isinstance(self.dataset, str):
            self.validate_dataset(self.dataset)

        if self.dataset is None or self.select_expression is None:
            raise InvalidSqlError(
                "Invalid statement - all statements require SELECT and FROM clauses."
            )


def SqlReader(sql_statement: str, **kwargs):
    """
    Use basic SQL queries to filter Reader.

    Parameters:
        sql_statement: string
        kwargs: parameters to pass to the Reader

    Note:
        `select` is taken from SQL SELECT
        `dataset` is taken from SQL FROM
        `filters` is taken from SQL WHERE
    """

    # some imports here to remove cyclic imports
    from mabel import DictSet
    from mabel import Reader

    sql = SqlParser(sql_statement)
    get_logger().info(repr(sql))

    actual_select = sql.select_expression
    if sql.select_expression is None:
        actual_select = "*"
    elif sql.select_expression != "*":
        actual_select = sql.select_expression + ", *"

    reducer = None
    if sql.select_expression == "COUNT(*)":
        reducer = lambda x: {"*": "*"}

    # FROM clause
    # WHERE clause
    if isinstance(sql.dataset, list):
        # it's a list if it's been parsed into a SQL statement,
        # this is how subqueries are interpretted - the parser
        # doesn't extract a dataset name - it collects parts of
        # a SQL statement which it can then pass to a SqlReader
        # to get back a dataset - which we then use as the
        # dataset for the outer query.
        reader = SqlReader("".join(sql.dataset), **kwargs)
    else:
        reader = Reader(
            select=actual_select,
            dataset=sql.dataset,
            filters=sql.where_expression,
            **kwargs,
        )

    # GROUP BY clause
    if sql.group_by or any(
        [t["type"] == TOKENS.AGGREGATOR for t in sql.select_evaluator.tokens]  # type:ignore
    ):
        from ...internals.group_by import GroupBy

        # convert the clause into something we can pass to GroupBy
        if sql.group_by:
            groups = [group.strip() for group in sql.group_by.split(",") if group.strip() != ""]
        else:
            groups = ["*"]  # we're not really grouping

        aggregations = []
        renames = []
        for t in sql.select_evaluator.tokens:  # type:ignore
            if t["type"] == TOKENS.AGGREGATOR:
                aggregations.append((t["value"], t["parameters"][0]["value"]))
                if t["as"]:
                    t["raw"] = get_function_name(t)
                    renames.append(t)
            elif t["type"] == TOKENS.VARIABLE and t["value"] not in groups:
                raise InvalidSqlError(
                    "Invalid SQL - SELECT clause in a statement with a GROUP BY clause must be made of aggregations or items from the GROUP BY clause."
                )

        if aggregations:
            grouped = GroupBy(reader, groups).aggregate(aggregations)
        else:
            grouped = GroupBy(reader, groups).groups()

        # there could be 250000 groups, so we're not going to load them into memory
        reader = DictSet(grouped)

    # HAVING clause
    # if we have a HAVING clause, filter the grouped data by it
    if sql.having:
        reader = reader.filter(sql.having)

    # SELECT clause
    renames = {}  # type:ignore
    for t in sql.select_evaluator.tokens:  # type:ignore
        if t["as"]:
            renames[get_function_name(t)] = t["as"]

    def _perform_renames(row):
        for k, v in [(k, v) for k, v in row.items()]:
            if k in renames:
                row[renames[k]] = row.pop(k, row.get(renames[k]))
        return row

    if renames:
        reader = DictSet(map(_perform_renames, reader))

    reader = reader.select(sql.select_evaluator.fields())  # type:ignore
    # distinct now we have only the columns we're interested in
    if sql.distinct:
        reader = reader.distinct()

    # ORDER BY clause
    if sql.order_by:
        take = 10000  # the Query UI is currently set to 2000
        if sql.limit:
            take = int(sql.limit)
        reader = DictSet(
            reader.sort_and_take(column=sql.order_by, take=take, descending=sql.order_descending)
        )

    # LIMIT clause
    if sql.limit:
        reader = reader.take(sql.limit)

    return reader



from . import BaseStorageClass


class StorageClassMemory(BaseStorageClass):
    """
    This provides the reader for the MEMORY variation of STORAGE.
    """

    def __init__(self, iterator):
        self.inner_reader = None
        self.data = list(iterator)
        self.length = len(self.data)
        self.iterator = None

    def _inner_reader(self, *locations):
        if locations:
            yield from [self.data[l] for l in locations]
        else:
            yield from self.data

    def __iter__(self):
        self.iterator = self._inner_reader()
        return self.iterator

    def __next__(self):
        if not self.iterator:
            self.iterator = self._inner_reader()
        return next(self.iterator)

    def __len__(self):
        return self.length



"""
We're testing the Expression module, but for simplicity we're testing it via
the .query() method on the DictSet, this makes handling the data we're querying
easier to handle.
"""
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data import STORAGE_CLASS
from mabel import DictSet
from mabel.data.internals.expression import Expression
from rich import traceback

traceback.install()

# fmt: off
TEST_DATA = [
    { "name": "Sirius Black", "age": 40, "alive": False, "dob": "1970-01-02", "gender": "male", "affiliations": ['OotP'] },
    { "name": "Harry Potter", "age": 11, "alive": True, "dob": "1999-07-30", "gender": "male", "affiliations": ['Dumbledores Army', 'Griffindor'] },
    { "name": "Hermione Grainger", "age": 10, "alive": True, "dob": "1999-12-14", "gender": "female", "affiliations": ['Dumbledores Army', 'Griffindor', 'MoM'] },
    { "name": "Fleur Isabelle Delacour", "age": 11, "alive": False, "dob": "1999-02-08", "gender": "female" },
    { "name": "James Potter", "age": 40, "alive": False, "dob": "1971-12-30", "gender": "male" },
    { "name": "James Potter", "age": 0, "alive": True, "dob": "2010-12-30", "gender": "male" },
    { "name": "Lily Potter", "age": 39, "alive": False, "dob": "1972-01-12", "gender": "female"}
]
# fmt: on


def test_expression_compilation():
    # fmt: off
    EXPRESSIONS = [
        {"expression": "YEAR(dob) == 1979", "dnf": ("YEAR(dob)", "==", 1979)},
        {"expression": "name == 'James Potter'","dnf": ("name", "==", '"James Potter"'),},
        {"expression": "not alive == True", "dnf": ("NOT", ("alive", "==", True))},
        {"expression": "age < 100 and alive == true","dnf": [("age", "<", 100), ("alive", "==", True)],},
        {"expression": "dob == '1979-09-19' or name like '%potter%'","dnf": [[("dob", "==", '"1979-09-19 00:00:00"')],[("name", "LIKE", '"%potter%"')],],},
        {"expression": "`name` == 'James Potter'", "dnf": ("`name`", "==", '"James Potter"')},
    ]
    # fmt: on

    for e in EXPRESSIONS:
        exp = Expression(e["expression"])
        dnf = exp.to_dnf()
        assert dnf == e["dnf"], dnf


def test_simple_equals_expressions():
    DATA = DictSet(TEST_DATA, storage_class=STORAGE_CLASS.MEMORY)

    assert DATA.filter("name == 'James Potter'").count() == 2
    assert DATA.filter("age == 10").count() == 1
    assert DATA.filter("alive == true").count() == 3
    assert DATA.filter("affiliations is none").count() == 4
    assert DATA.filter("name like '%Potter%'").count() == 4


def test_simple_not_expressions():
    DATA = DictSet(TEST_DATA, storage_class=STORAGE_CLASS.MEMORY)

    assert DATA.filter("name <> 'James Potter'").count() == 5
    assert DATA.filter("not age == 10").count() == 6
    assert DATA.filter("alive == false").count() == 4
    assert DATA.filter("affiliations is not none").count() == 3
    assert DATA.filter("name not like '%Potter%'").count() == 3


def test_simple_compound_expressions():
    DATA = DictSet(TEST_DATA, storage_class=STORAGE_CLASS.MEMORY)

    assert DATA.filter("name like '%Potter' and alive == true").count() == 2
    assert DATA.filter("name like '%Potter' or alive == true").count() == 5


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import datetime
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data import Reader
from rich import traceback
from mabel.data.readers.internals import inline_functions

traceback.install()


FIXED_DATE = datetime.datetime(2000, 6, 12, 8, 43, 22)

print(FIXED_DATE)


def test_inline_date_functions():
    assert inline_functions.get_year("2021-01-02") == 2021
    assert inline_functions.get_year("2021-01-02T00:00") == 2021
    assert inline_functions.get_year(FIXED_DATE) == 2000

    assert inline_functions.get_month("2021-01-02") == 1
    assert inline_functions.get_month("2021-01-02T00:00") == 1
    assert inline_functions.get_month(FIXED_DATE) == 6

    assert inline_functions.get_day("2021-01-02") == 2
    assert inline_functions.get_day("2021-01-02T00:00") == 2
    assert inline_functions.get_day(FIXED_DATE) == 12

    assert inline_functions.get_date("2021-01-02") == datetime.date(2021, 1, 2)
    assert inline_functions.get_date("2021-01-02T00:00") == datetime.date(2021, 1, 2)
    assert inline_functions.get_date(FIXED_DATE) == datetime.date(2000, 6, 12)

    assert inline_functions.get_time("2021-01-02T10:12:03") == datetime.time(10, 12, 3)
    assert inline_functions.get_time("2021-01-02T00:00") == datetime.time(0, 0, 0)
    assert inline_functions.get_time(FIXED_DATE) == datetime.time(8, 43, 22)

    assert inline_functions.get_quarter("2021-01-02T10:12:03") == 1
    assert inline_functions.get_quarter("2021-01-02T00:00") == 1
    assert inline_functions.get_quarter(FIXED_DATE) == 2

    assert inline_functions.get_hour("2021-01-02T10:12:03") == 10
    assert inline_functions.get_hour("2021-01-02T00:00") == 0
    assert inline_functions.get_hour(FIXED_DATE) == 8

    assert inline_functions.get_minute("2021-01-02T10:12:03") == 12
    assert inline_functions.get_minute("2021-01-02T00:00") == 0
    assert inline_functions.get_minute(FIXED_DATE) == 43

    assert inline_functions.get_second("2021-01-02T10:12:03") == 3
    assert inline_functions.get_second("2021-01-02T00:00") == 0
    assert inline_functions.get_second(FIXED_DATE) == 22

    assert inline_functions.get_week("2021-01-10T10:12:03") == 1
    assert inline_functions.get_week("2021-01-02T00:00") == 53
    assert inline_functions.get_week(FIXED_DATE) == 24, inline_functions.get_week(FIXED_DATE)

    assert inline_functions.add_days("2021-01-10T10:12:03", 10) == datetime.datetime(
        2021, 1, 20, 10, 12, 3
    )
    assert inline_functions.add_days("2021-01-02T00:00", -10) == datetime.datetime(
        2020, 12, 23, 0, 0
    )
    assert inline_functions.add_days(FIXED_DATE, 10) == datetime.datetime(2000, 6, 22, 8, 43, 22)

    assert inline_functions.diff_days("2021-01-10T10:12:03", FIXED_DATE) == -7518


def test_inline_other():
    vals = []
    for i in range(10):
        vals.append(inline_functions.get_random())
    vals = set(vals)
    assert 9 <= len(vals) <= 10, len(vals)

    assert inline_functions.concat("a", "b", "c") == "abc"
    assert inline_functions.concat(["a", "b", "c"]) == "a, b, c"

    assert inline_functions.get_md5("a") == "0cc175b9c0f1b6a831c399e269772661"

    assert inline_functions.to_string({"apples": 2, "pears": 4}) == '\\{"apples":2,"pears":4}\\'


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



from setuptools import find_packages
from setuptools import setup

with open("mabel/version.py", "r") as v:
    vers = v.read()
exec(vers)  # nosec

with open("README.md", "r") as rm:
    long_description = rm.read()

try:
    with open("requirements.txt", "r") as f:
        required = f.read().splitlines()
except:
    with open("mabel.egg-info/requires.txt", "r") as f:
        required = f.read().splitlines()

setup(
    name="mabel",
    version=__version__,
    description="Python Data Libraries",
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Joocer",
    author="joocer",
    author_email="justin.joyce@joocer.com",
    packages=find_packages(include=["mabel", "mabel.*"]),
    url="https://github.com/mabel-dev/mabel/",
    install_requires=required,
)



"""
We're testing the Expression module, but for simplicity we're testing it via
the .query() method on the DictSet, this makes handling the data we're querying
easier to handle.
"""

import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from rich import traceback
from mabel.data.readers.internals.inline_evaluator import Evaluator

traceback.install()


TEST_DATA = {
    "name": "Sirius Black",
    "age": 40,
    "alive": False,
    "dob": "1970-01-02",
    "gender": "male",
    "affiliations": ["OotP"],
}


def test_simple_fields():
    pf = "name, age, alive, dob, gender"
    res = Evaluator(pf)(TEST_DATA)
    assert res == {
        "name": "Sirius Black",
        "age": 40,
        "alive": False,
        "dob": "1970-01-02",
        "gender": "male",
    }, res


def test_simple_functions():
    pf = "UPPER(name), YEAR(dob), LEFT(gender, 2), RIGHT(gender, 2), MID(name, 5, 3)"
    res = Evaluator(pf)(TEST_DATA)
    assert res == {
        "UPPER(name)": "SIRIUS BLACK",
        "YEAR(dob)": 1970,
        "LEFT(gender,2)": "ma",
        "RIGHT(gender,2)": "le",
        "MID(name,5,3)": "s B",
    }, res


def test_concat():
    pf = "CONCAT(name,' ',name,' table')"
    res = Evaluator(pf)(TEST_DATA)
    assert res == {"CONCAT(name,' ',name,' table')": "Sirius Black Sirius Black table"}, res


def test_as():
    pf = "UPPER(name) as upper_name"
    res = Evaluator(pf)(TEST_DATA)
    assert res == {"upper_name": "SIRIUS BLACK"}


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
import mabel
from rich import traceback

traceback.install()


def test_version():
    assert hasattr(mabel, "__version__")


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
Multiprocessing is not faster in benchmarks, this is being retained but will need
to be manually enabled.

When a reliable use case for multiprocessing is identified it may be included into the
automatic running of the data accesses.
"""
import datetime
import logging
import multiprocessing
import os
import time
from multiprocessing import Queue
from queue import Empty

import lz4.frame

from .parsers import json

TERMINATE_SIGNAL = -1
MAXIMUM_SECONDS_PROCESSES_CAN_RUN = 600


def fix_dict(obj: dict) -> dict:
    def fix_fields(dt):
        dt_type = type(dt)
        if dt_type in (int, float, bool, str):
            return dt
        if hasattr(dt, "mini"):
            return dt.mini.decode("UTF8")
        if dt_type == dict:
            return str(fix_dict(dt))
        if dt_type in (list, tuple, set):
            return str([fix_fields(i) for i in dt])
        if dt_type in (datetime.date, datetime.datetime):
            return dt.isoformat()
        if dt_type == bytes:
            return dt.decode("UTF8")
        return str(dt)

    if not isinstance(obj, dict):
        return obj  # type:ignore

    for key in obj.keys():
        obj[key] = fix_fields(obj[key])
    return obj


def serialize(ob):
    import orjson

    return orjson.dumps(fix_dict(ob))


def _inner_process(func, source_queue, reply_queue):  # pragma: no cover
    try:
        source = source_queue.get(timeout=1)
    except Empty:  # pragma: no cover
        source = TERMINATE_SIGNAL

    while source != TERMINATE_SIGNAL:
        # non blocking wait - this isn't thread aware in that it can trivially
        # have race conditions, but it will apply a simple back-off so we're
        # not exhausting memory when we know we should wait
        while reply_queue.full():
            time.sleep(1)
        # the empty list here is where the list of indicies should go
        reply_queue.put(
            lz4.frame.compress(b"\n".join(map(serialize, func(source, [])))),
            timeout=30,
        )
        source = None
        while source is None:
            try:
                source = source_queue.get(timeout=1)
            except Empty:  # pragma: no cover
                source = None


def processed_reader(func, items_to_read, support_files):  # pragma: no cover
    if os.name == "nt":  # pragma: no cover
        raise NotImplementedError("Reader Multi Processing not available on Windows platforms")

    process_pool = []

    # determine the number of slots we're going to make available:
    # - less than or equal to the number of files to read
    # - one less than the CPUs we have
    slots = max(min(len(items_to_read), multiprocessing.cpu_count() - 1), 2)
    reply_queue = Queue(slots)

    send_queue = multiprocessing.Queue()
    for item_index in range(slots):
        if item_index < len(items_to_read):
            send_queue.put(items_to_read[item_index])

    # We're going to use all but one CPU, unless there's 1 or 2 CPUs, then we're going
    # to create two processes
    for i in range(slots):
        process = multiprocessing.Process(
            target=_inner_process,
            args=(func, send_queue, reply_queue),
        )
        process.daemon = True
        process.start()
        process_pool.append(process)

    process_start_time = time.time()
    item_index = slots

    while (
        any({p.is_alive() for p in process_pool})
        or not reply_queue.empty()
        or not send_queue.empty()
    ):
        try:
            records = reply_queue.get(timeout=1)
            yield from map(json, [r for r in lz4.frame.decompress(records).split(b"\n") if r])
            if item_index < len(items_to_read):
                send_queue.put_nowait(items_to_read[item_index])
                item_index += 1
            else:
                send_queue.put_nowait(TERMINATE_SIGNAL)

        except Empty:  # nosec
            if time.time() - process_start_time > MAXIMUM_SECONDS_PROCESSES_CAN_RUN:
                logging.error(
                    f"Sending TERMINATE to long running multi-processed processes after {MAXIMUM_SECONDS_PROCESSES_CAN_RUN} seconds total run time"
                )
                break
        except GeneratorExit:
            logging.error("GENERATOR EXIT DETECTED")
            break

    reply_queue.close()
    send_queue.close()
    reply_queue.join_thread()
    send_queue.join_thread()
    for process in process_pool:
        process.join()



import time
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.disk import DiskWriter, DiskReader
from mabel.data import StreamWriter
from mabel.data import Reader
from mabel.data.validator import schema_loader
import shutil
from pathlib import Path
from rich import traceback

traceback.install()


DATA_SET = [
    {"key": 6},
    {"key": 10},
    {"key": 3},
    {"key": 9},
    {"key": "eight"},
    {"key": 4},
    {"key": 7},
    {"key": 5},
    {"key": "two"},
    {"key": 1},
]
SCHEMA = {"fields": [{"name": "key", "type": "INTEGER"}]}
TEST_FOLDER = "_temp/path"


def test_writer_backout():
    if Path(TEST_FOLDER).exists():  # pragma: no cover
        shutil.rmtree(TEST_FOLDER)

    w = StreamWriter(
        dataset=TEST_FOLDER,
        inner_writer=DiskWriter,
        schema=SCHEMA,
        idle_timeout_seconds=1,
    )

    for record in DATA_SET:
        w.append(record)

    time.sleep(4)

    r = Reader(dataset=TEST_FOLDER, inner_reader=DiskReader)

    assert len(list(r)) == 8


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
There's nothing clever here, it's just a dumb iterator wrapper
"""


class DumbIterator:
    def __init__(self, i):
        self._i = iter(i)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._i)



"""
Note that this is testing randomness functions, failure is inevitable on a
long enough series of tests (think billions)
"""
import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.utils import entropy
from rich import traceback

traceback.install()


def test_random_string():
    """test the string is the right length"""
    rnd_str = entropy.random_string(length=32)
    assert len(rnd_str) == 32


def test_random_int():
    # this test will eventually fail
    rnd_1 = entropy.random_int()
    rnd_2 = entropy.random_int()

    assert not rnd_1 == rnd_2
    assert isinstance(rnd_1, int)


def test_random_range():
    vals = []
    for i in range(5000):
        vals.append(entropy.random_range(10, 20))

    assert min(vals) == 10, "test lower extreme of range"
    assert max(vals) == 20, "test upper extreme of range"


def test_bytes_to_int():
    b1 = [0] * 8
    assert entropy.bytes_to_int(b1) == 0

    b2 = [255] * 4
    assert entropy.bytes_to_int(b2) == (2**32 - 1)

    b3 = [255] * 8
    assert entropy.bytes_to_int(b3) == (2**64 - 1)


def test_random_choice():
    options = ["one", "two", "three"]

    for i in range(100):
        choice = entropy.random_choice(options)
        assert choice in options


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
XML helper function, we don't create XML data, we read it.
"""

from collections import defaultdict
from xml.etree import cElementTree as ElementTree  # nosec


def strip_namespace(entry):
    if isinstance(entry, dict):
        for k in [k for k in entry.keys() if k.startswith("{")]:
            k2 = k.split("}", 1)[1]
            entry[k2] = entry.pop(k)
        for k in [k for k in entry.keys() if k.startswith("@{")]:
            k2 = k.split("}", 1)[1]
            entry[f"@{k2}"] = entry.pop(k)
        for child in entry:
            if isinstance(entry[child], (list, dict)):
                strip_namespace(entry[child])
    if isinstance(entry, list):
        for child in entry:
            if isinstance(child, (list, dict)):
                strip_namespace(child)


def etree_to_dict(t):
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[t.tag].update(("@" + k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]["#text"] = text
        else:
            d[t.tag] = text
    return d


def parse(xml_string):
    tree = ElementTree.XML(xml_string)
    dictionary = etree_to_dict(tree)
    strip_namespace(dictionary)
    return dictionary



"""
Separate the implementation of the Cursor from the Reader.

Cursor is made of three parts:
- map      : a bit array representing all of the blobs in the set - unread blobs
             are 0s and read blobs are 1s. This allows for blobs to be read in 
             an arbitrary order - although currently only implemented linearly.
- partition: the active parition (blob) that is being read
- location : the record in the active partition (blob), so we can resume reading
             midway through the blob if required.
"""
import orjson
from orso.bitarray import BitArray
from orso.cityhash import CityHash64


class InvalidCursor(Exception):
    pass


class Cursor:
    def __init__(self, readable_blobs, cursor=None):
        # sort the readable blobs so they are in a consistent order
        self.readable_blobs = sorted(readable_blobs)
        self.read_blobs = []
        self.partition = ""
        self.location = -1

        if cursor:
            self.load_cursor(cursor)

    def load_cursor(self, cursor):
        if cursor is None:
            return

        if isinstance(cursor, str):
            cursor = orjson.loads(cursor)

        if (
            not "location" in cursor.keys()
            or not "map" in cursor.keys()
            or not "partition" in cursor.keys()
        ):
            raise InvalidCursor(f"Cursor is malformed or corrupted {cursor}")

        self.location = cursor["location"]
        find_partition = [
            blob for blob in self.readable_blobs if CityHash64(blob) == cursor["partition"]
        ]
        if len(find_partition) == 1:
            self.partition = find_partition[0]
        map_bytes = bytes.fromhex(cursor["map"])
        blob_map = BitArray.from_array(map_bytes, len(map_bytes) * 8)
        self.read_blobs = [
            self.readable_blobs[i] for i in range(len(self.readable_blobs)) if blob_map.get(i)
        ]

    def next_blob(self, previous_blob=None):
        if previous_blob:
            self.read_blobs.append(previous_blob)
            self.partition = ""
            self.location = -1
        if self.partition and self.location >= 0:
            if self.partition in self.readable_blobs:
                return self.partition
            partition_finder = [
                blob for blob in self.readable_blobs if CityHash64(blob) == self.partition
            ]
            if len(partition_finder) != 1:
                raise ValueError(f"Unable to determine current partition ({self.partition})")
            return partition_finder[0]
        unread = [blob for blob in self.readable_blobs if blob not in self.read_blobs]
        if len(unread) > 0:
            self.partition = unread[0]
            self.location = -1
            return self.partition
        return None

    def skip_to_cursor(self, iterator):
        # cycle through the iterator to the cursor location
        if self.location < 0:
            return 0
        for index in range(self.location + 1):
            next(iterator, None)
        return self.location + 1

    def get(self):
        return {
            "map": self["map"],
            "partition": self["partition"],
            "location": self["location"],
        }

    def __getitem__(self, item):
        if item == "map":
            blob_map = BitArray(len(self.readable_blobs))
            for i, blob in enumerate(self.readable_blobs):
                blob_map.set(i, blob in self.read_blobs)
            return blob_map.array.hex()
        if item == "partition":
            return CityHash64(self.partition)
        if item == "location":
            return self.location
        return None

    def __repr__(self):
        return orjson.dumps(self.get()).decode("UTF8")



import shutil
import os
import sys
import glob

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.disk import DiskReader, DiskWriter
from mabel.adapters.null import NullWriter
from mabel.data import Reader, Writer


def do_writer():
    shutil.rmtree("_temp", ignore_errors=True)
    w = Writer(inner_writer=DiskWriter, dataset="_temp", schema=["name", "alter"])
    for i in range(int(1e5)):
        w.append({"name": "Barney Stinson", "alter": "Lorenzo Von Matterhorn"})
        w.append({"name": "Laszlo Cravensworth", "alter": "Jackie Daytona"})
        w.append({"name": "Pheobe Buffay", "alter": "Regina Phalange"})
    w.finalize()


def do_writer_compressed(algo):
    w = Writer(inner_writer=DiskWriter, dataset="_temp", format=algo, schema=["test"])
    for i in range(int(1e5)):
        w.append({"test": True})
        w.append({"test": False})
    w.finalize()
    del w


def do_writer_default():
    w = Writer(inner_writer=DiskWriter, dataset="_temp", schema=["name", "person"])
    for i in range(int(1e5)):
        w.append({"name": "Barney Stinson", "person": "Lorenzo Von Matterhorn"})
        w.append({"name": "Laszlo Cravensworth", "person": "Jackie Daytona"})
    w.finalize()
    del w


def do_writer_csv():
    w = Writer(inner_writer=DiskWriter, dataset="_temp", format="text", schema=[])
    w.append("week,volume")
    for i in range(int(100)):
        w.append(f"{i},{i}")
    w.finalize()


def test_reader_writer():
    do_writer()
    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 300000, l


def test_reader_writer_format_zstd():
    do_writer_compressed("zstd")
    g = glob.glob("_temp/**/*.zstd", recursive=True)
    assert len(g) > 0, g
    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_format_jsonl():
    do_writer_compressed("jsonl")
    g = glob.glob("_temp/**/*.jsonl", recursive=True)
    assert len(g) > 0, g

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_format_parquet():
    do_writer_compressed("parquet")
    g = glob.glob("_temp/**/*.parquet", recursive=True)
    assert len(g) > 0, g
    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_reader_writer_format_text():
    do_writer_compressed("text")
    g = glob.glob("_temp/**/*.txt", recursive=True)
    assert len(g) > 0, g

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 1, l


def test_reader_writer_format_default():
    do_writer_default()
    g = glob.glob("_temp/**/*.zstd", recursive=True)
    assert len(g) > 0, g

    r = Reader(inner_reader=DiskReader, dataset="_temp")
    l = len(list(r))
    shutil.rmtree("_temp", ignore_errors=True)
    assert l == 200000, l


def test_write_to_path_logged():
    # none of these should do anything
    nw = Writer(inner_writer=NullWriter, to_path="bucket/path", schema=["abc"])
    nw.append({"abc": "def"})
    print(nw.finalize())


def get_data():
    r = Reader(inner_reader=DiskReader, dataset="tests/data/tweets", raw_path=True)
    return r


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import time
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.null import NullWriter
from mabel.data import StreamWriter
from rich import traceback
from freezegun import freeze_time

traceback.install()


def test_stream_rollover():
    # none of these should do anything
    w = StreamWriter(
        dataset="bucket/path/file.extension",
        inner_writer=NullWriter,
        idle_timeout_seconds=1,
        format="text",
        schema=["@"],
    )

    with freeze_time("2012-01-14"):
        lines = w.append("It's 2012")
        lines = w.append("It's still 2012")
        assert lines == 2

    # when we change the date, we write to a new partition
    with freeze_time("2017-01-14"):
        lines = w.append("Now it's 2017")
        assert lines == 1
        lines = w.append("It's still 2017")
        assert lines == 2

    time.sleep(2)


def test_fixed_dates():
    # none of these should do anything
    w = StreamWriter(
        dataset="bucket/path/file.extension",
        inner_writer=NullWriter,
        idle_timeout_seconds=10,
        format="text",
        date="2022-01-01",
        schema=["@"],
    )

    # when we fix the date (use the data param), we always write to the same date
    with freeze_time("2012-01-14"):
        lines = w.append("It's 2012")
        lines = w.append("It's still 2012")
        assert lines == 2

    with freeze_time("2017-01-14"):
        lines = w.append("Now it's 2017")
        assert lines == 3
        lines = w.append("It's still 2017")
        assert lines == 4

    time.sleep(2)


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()
